# -*- coding: utf-8 -*-
#   Copyright 2018 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Tests for projectq.cengines._linearmapper.py."""
from copy import deepcopy

import pytest

from projectq.cengines import DummyEngine
from projectq.meta import LogicalQubitIDTag
from projectq.ops import (
    Allocate,
    BasicGate,
    CNOT,
    Command,
    Deallocate,
    FlushGate,
    QFT,
    X,
)
from projectq.types import WeakQubitRef

from projectq.cengines import _linearmapper as lm


def test_return_swap_depth():
    swaps = []
    assert lm.return_swap_depth(swaps) == 0
    swaps += [(0, 1), (0, 1), (1, 2)]
    assert lm.return_swap_depth(swaps) == 3
    swaps.append((2, 3))
    assert lm.return_swap_depth(swaps) == 4


def test_is_available():
    mapper = lm.LinearMapper(num_qubits=5, cyclic=False)
    qb0 = WeakQubitRef(engine=None, idx=0)
    qb1 = WeakQubitRef(engine=None, idx=1)
    qb2 = WeakQubitRef(engine=None, idx=2)
    cmd0 = Command(None, BasicGate(), qubits=([qb0],))
    assert mapper.is_available(cmd0)
    cmd1 = Command(None, BasicGate(), qubits=([qb0],), controls=[qb1])
    assert mapper.is_available(cmd1)
    cmd2 = Command(None, BasicGate(), qubits=([qb0], [qb1, qb2]))
    assert not mapper.is_available(cmd2)
    cmd3 = Command(None, BasicGate(), qubits=([qb0], [qb1]), controls=[qb2])
    assert not mapper.is_available(cmd3)


def test_return_new_mapping_too_many_qubits():
    mapper = lm.LinearMapper(num_qubits=3, cyclic=False)
    qb0 = WeakQubitRef(engine=None, idx=0)
    qb1 = WeakQubitRef(engine=None, idx=1)
    qb2 = WeakQubitRef(engine=None, idx=2)
    cmd0 = Command(None, QFT, qubits=([qb0], [qb1, qb2]))
    mapper._stored_commands = [cmd0]
    with pytest.raises(Exception):
        mapper.return_new_mapping(
            num_qubits=mapper.num_qubits,
            cyclic=mapper.cyclic,
            currently_allocated_ids=mapper._currently_allocated_ids,
            stored_commands=mapper._stored_commands,
            current_mapping=mapper.current_mapping,
        )
    cmd1 = Command(None, BasicGate(), qubits=([],))
    mapper._stored_commands = [cmd1]
    with pytest.raises(Exception):
        mapper.return_new_mapping(
            num_qubits=mapper.num_qubits,
            cyclic=mapper.cyclic,
            currently_allocated_ids=mapper._currently_allocated_ids,
            stored_commands=mapper._stored_commands,
            current_mapping=mapper.current_mapping,
        )


def test_return_new_mapping_allocate_qubits():
    mapper = lm.LinearMapper(num_qubits=2, cyclic=False)
    qb0 = WeakQubitRef(engine=None, idx=0)
    qb1 = WeakQubitRef(engine=None, idx=1)
    mapper._currently_allocated_ids = set([4])
    cmd0 = Command(None, Allocate, ([qb0],))
    cmd1 = Command(None, Allocate, ([qb1],))
    mapper._stored_commands = [cmd0, cmd1]
    new_mapping = mapper.return_new_mapping(
        num_qubits=mapper.num_qubits,
        cyclic=mapper.cyclic,
        currently_allocated_ids=mapper._currently_allocated_ids,
        stored_commands=mapper._stored_commands,
        current_mapping=mapper.current_mapping,
    )
    assert mapper._currently_allocated_ids == set([4])
    assert mapper._stored_commands == [cmd0, cmd1]
    assert len(new_mapping) == 2
    assert 4 in new_mapping and 0 in new_mapping


def test_return_new_mapping_allocate_only_once():
    mapper = lm.LinearMapper(num_qubits=1, cyclic=False)
    qb0 = WeakQubitRef(engine=None, idx=0)
    qb1 = WeakQubitRef(engine=None, idx=1)  # noqa: F841
    mapper._currently_allocated_ids = set()
    cmd0 = Command(None, Allocate, ([qb0],))
    cmd1 = Command(None, Deallocate, ([qb0],))
    # Test if loop stops after deallocate gate has been used.
    # This would otherwise trigger an error (test by num_qubits=2)
    cmd2 = None
    mapper._stored_commands = [cmd0, cmd1, cmd2]
    mapper.return_new_mapping(
        num_qubits=mapper.num_qubits,
        cyclic=mapper.cyclic,
        currently_allocated_ids=mapper._currently_allocated_ids,
        stored_commands=mapper._stored_commands,
        current_mapping=mapper.current_mapping,
    )


def test_return_new_mapping_possible_map():
    mapper = lm.LinearMapper(num_qubits=3, cyclic=False)
    qb0 = WeakQubitRef(engine=None, idx=0)
    qb1 = WeakQubitRef(engine=None, idx=1)
    qb2 = WeakQubitRef(engine=None, idx=2)
    cmd0 = Command(None, Allocate, ([qb0],))
    cmd1 = Command(None, Allocate, ([qb1],))
    cmd2 = Command(None, Allocate, ([qb2],))
    cmd3 = Command(None, CNOT, qubits=([qb0],), controls=[qb1])
    cmd4 = Command(None, CNOT, qubits=([qb2],), controls=[qb1])
    cmd5 = Command(None, X, qubits=([qb0],))
    mapper._stored_commands = [cmd0, cmd1, cmd2, cmd3, cmd4, cmd5]
    new_mapping = mapper.return_new_mapping(
        num_qubits=mapper.num_qubits,
        cyclic=mapper.cyclic,
        currently_allocated_ids=mapper._currently_allocated_ids,
        stored_commands=mapper._stored_commands,
        current_mapping=mapper.current_mapping,
    )
    assert new_mapping == {0: 2, 1: 1, 2: 0} or new_mapping == {0: 0, 1: 1, 2: 2}


def test_return_new_mapping_previous_error():
    mapper = lm.LinearMapper(num_qubits=2, cyclic=False)
    qb0 = WeakQubitRef(engine=None, idx=0)
    qb1 = WeakQubitRef(engine=None, idx=1)
    qb2 = WeakQubitRef(engine=None, idx=2)
    qb3 = WeakQubitRef(engine=None, idx=3)
    cmd0 = Command(None, Allocate, ([qb0],))
    cmd1 = Command(None, Allocate, ([qb1],))
    cmd2 = Command(None, Allocate, ([qb2],))
    cmd3 = Command(None, Allocate, ([qb3],))
    cmd4 = Command(None, CNOT, qubits=([qb2],), controls=[qb3])
    mapper._stored_commands = [cmd0, cmd1, cmd2, cmd3, cmd4]
    mapper.return_new_mapping(
        num_qubits=mapper.num_qubits,
        cyclic=mapper.cyclic,
        currently_allocated_ids=mapper._currently_allocated_ids,
        stored_commands=mapper._stored_commands,
        current_mapping=mapper.current_mapping,
    )


def test_process_two_qubit_gate_not_in_segments_test0():
    mapper = lm.LinearMapper(num_qubits=5, cyclic=False)
    segments = [[0, 1]]
    active_qubits = set([0, 1, 4, 6])
    neighbour_ids = {0: set([1]), 1: set([0]), 4: set(), 6: set()}
    mapper._process_two_qubit_gate(
        num_qubits=mapper.num_qubits,
        cyclic=mapper.cyclic,
        qubit0=4,
        qubit1=6,
        active_qubits=active_qubits,
        segments=segments,
        neighbour_ids=neighbour_ids,
    )
    assert len(segments) == 2
    assert segments[0] == [0, 1]
    assert segments[1] == [4, 6]
    assert neighbour_ids[4] == set([6])
    assert neighbour_ids[6] == set([4])
    assert active_qubits == set([0, 1, 4, 6])


def test_process_two_qubit_gate_not_in_segments_test1():
    mapper = lm.LinearMapper(num_qubits=5, cyclic=False)
    segments = []
    active_qubits = set([4, 6])
    neighbour_ids = {4: set(), 6: set()}
    mapper._process_two_qubit_gate(
        num_qubits=mapper.num_qubits,
        cyclic=mapper.cyclic,
        qubit0=5,
        qubit1=6,
        active_qubits=active_qubits,
        segments=segments,
        neighbour_ids=neighbour_ids,
    )
    assert len(segments) == 0
    assert active_qubits == set([4])


@pytest.mark.parametrize("qb0, qb1", [(1, 2), (2, 1)])
def test_process_two_qubit_gate_one_qb_free_one_qb_in_segment(qb0, qb1):
    # add on the right to segment
    mapper = lm.LinearMapper(num_qubits=3, cyclic=False)
    segments = [[0, 1]]
    active_qubits = set([0, 1, 2])
    neighbour_ids = {0: set([1]), 1: set([0]), 2: set()}
    mapper._process_two_qubit_gate(
        num_qubits=mapper.num_qubits,
        cyclic=mapper.cyclic,
        qubit0=qb0,
        qubit1=qb1,
        active_qubits=active_qubits,
        segments=segments,
        neighbour_ids=neighbour_ids,
    )
    assert segments == [[0, 1, 2]]
    assert active_qubits == set([0, 1, 2])
    assert neighbour_ids[1] == set([0, 2])
    assert neighbour_ids[2] == set([1])


@pytest.mark.parametrize("qb0, qb1", [(0, 1), (1, 0)])
def test_process_two_qubit_gate_one_qb_free_one_qb_in_segment2(qb0, qb1):
    # add on the left to segment
    mapper = lm.LinearMapper(num_qubits=3, cyclic=False)
    segments = [[1, 2]]
    active_qubits = set([0, 1, 2])
    neighbour_ids = {0: set([]), 1: set([2]), 2: set([1])}
    mapper._process_two_qubit_gate(
        num_qubits=mapper.num_qubits,
        cyclic=mapper.cyclic,
        qubit0=qb0,
        qubit1=qb1,
        active_qubits=active_qubits,
        segments=segments,
        neighbour_ids=neighbour_ids,
    )
    assert segments == [[0, 1, 2]]
    assert active_qubits == set([0, 1, 2])
    assert neighbour_ids[1] == set([0, 2])
    assert neighbour_ids[0] == set([1])


@pytest.mark.parametrize("qb0, qb1", [(1, 2), (2, 1)])
def test_process_two_qubit_gate_one_qb_free_one_qb_in_segment_cycle(qb0, qb1):
    mapper = lm.LinearMapper(num_qubits=3, cyclic=True)
    segments = [[0, 1]]
    active_qubits = set([0, 1, 2])
    neighbour_ids = {0: set([1]), 1: set([0]), 2: set()}
    mapper._process_two_qubit_gate(
        num_qubits=mapper.num_qubits,
        cyclic=mapper.cyclic,
        qubit0=qb0,
        qubit1=qb1,
        active_qubits=active_qubits,
        segments=segments,
        neighbour_ids=neighbour_ids,
    )
    assert segments == [[0, 1, 2]]
    assert active_qubits == set([0, 1, 2])
    assert neighbour_ids[1] == set([0, 2])
    assert neighbour_ids[2] == set([1, 0])


@pytest.mark.parametrize("qb0, qb1", [(1, 2), (2, 1)])
def test_process_two_qubit_gate_one_qb_free_one_qb_in_seg_cycle2(qb0, qb1):
    # not yet long enough segment for cycle
    mapper = lm.LinearMapper(num_qubits=4, cyclic=True)
    segments = [[0, 1]]
    active_qubits = set([0, 1, 2])
    neighbour_ids = {0: set([1]), 1: set([0]), 2: set()}
    mapper._process_two_qubit_gate(
        num_qubits=mapper.num_qubits,
        cyclic=mapper.cyclic,
        qubit0=qb0,
        qubit1=qb1,
        active_qubits=active_qubits,
        segments=segments,
        neighbour_ids=neighbour_ids,
    )
    assert segments == [[0, 1, 2]]
    assert active_qubits == set([0, 1, 2])
    assert neighbour_ids[1] == set([0, 2])
    assert neighbour_ids[2] == set([1])


def test_process_two_qubit_gate_one_qubit_in_middle_of_segment():
    mapper = lm.LinearMapper(num_qubits=5, cyclic=False)
    segments = []
    active_qubits = set([0, 1, 2, 3])
    neighbour_ids = {0: set([1]), 1: set([0, 2]), 2: set([1]), 3: set()}
    mapper._process_two_qubit_gate(
        num_qubits=mapper.num_qubits,
        cyclic=mapper.cyclic,
        qubit0=1,
        qubit1=3,
        active_qubits=active_qubits,
        segments=segments,
        neighbour_ids=neighbour_ids,
    )
    assert len(segments) == 0
    assert active_qubits == set([0, 2])


def test_process_two_qubit_gate_both_in_same_segment():
    mapper = lm.LinearMapper(num_qubits=3, cyclic=False)
    segments = [[0, 1, 2]]
    active_qubits = set([0, 1, 2])
    neighbour_ids = {0: set([1]), 1: set([0, 2]), 2: set([1])}
    mapper._process_two_qubit_gate(
        num_qubits=mapper.num_qubits,
        cyclic=mapper.cyclic,
        qubit0=0,
        qubit1=2,
        active_qubits=active_qubits,
        segments=segments,
        neighbour_ids=neighbour_ids,
    )
    assert segments == [[0, 1, 2]]
    assert active_qubits == set([1])


def test_process_two_qubit_gate_already_connected():
    mapper = lm.LinearMapper(num_qubits=3, cyclic=False)
    segments = [[0, 1, 2]]
    active_qubits = set([0, 1, 2])
    neighbour_ids = {0: set([1]), 1: set([0, 2]), 2: set([1])}
    mapper._process_two_qubit_gate(
        num_qubits=mapper.num_qubits,
        cyclic=mapper.cyclic,
        qubit0=0,
        qubit1=1,
        active_qubits=active_qubits,
        segments=segments,
        neighbour_ids=neighbour_ids,
    )
    assert segments == [[0, 1, 2]]
    assert active_qubits == set([0, 1, 2])


@pytest.mark.parametrize(
    "qb0, qb1, result_seg",
    [
        (0, 2, [1, 0, 2, 3]),
        (0, 3, [2, 3, 0, 1]),
        (1, 2, [0, 1, 2, 3]),
        (1, 3, [0, 1, 3, 2]),
    ],
)
def test_process_two_qubit_gate_combine_segments(qb0, qb1, result_seg):
    mapper = lm.LinearMapper(num_qubits=4, cyclic=False)
    segments = [[0, 1], [2, 3]]
    active_qubits = set([0, 1, 2, 3, 4])
    neighbour_ids = {0: set([1]), 1: set([0]), 2: set([3]), 3: set([2])}
    mapper._process_two_qubit_gate(
        num_qubits=mapper.num_qubits,
        cyclic=mapper.cyclic,
        qubit0=qb0,
        qubit1=qb1,
        active_qubits=active_qubits,
        segments=segments,
        neighbour_ids=neighbour_ids,
    )
    assert segments == [result_seg] or segments == [reversed(result_seg)]
    assert qb1 in neighbour_ids[qb0]
    assert qb0 in neighbour_ids[qb1]


@pytest.mark.parametrize(
    "qb0, qb1, result_seg",
    [
        (0, 2, [1, 0, 2, 3]),
        (0, 3, [2, 3, 0, 1]),
        (1, 2, [0, 1, 2, 3]),
        (1, 3, [0, 1, 3, 2]),
    ],
)
def test_process_two_qubit_gate_combine_segments_cycle(qb0, qb1, result_seg):
    mapper = lm.LinearMapper(num_qubits=4, cyclic=True)
    segments = [[0, 1], [2, 3]]
    active_qubits = set([0, 1, 2, 3, 4])
    neighbour_ids = {0: set([1]), 1: set([0]), 2: set([3]), 3: set([2])}
    mapper._process_two_qubit_gate(
        num_qubits=mapper.num_qubits,
        cyclic=mapper.cyclic,
        qubit0=qb0,
        qubit1=qb1,
        active_qubits=active_qubits,
        segments=segments,
        neighbour_ids=neighbour_ids,
    )
    assert segments == [result_seg] or segments == [reversed(result_seg)]
    assert qb1 in neighbour_ids[qb0]
    assert qb0 in neighbour_ids[qb1]
    assert result_seg[0] in neighbour_ids[result_seg[-1]]
    assert result_seg[-1] in neighbour_ids[result_seg[0]]


@pytest.mark.parametrize(
    "qb0, qb1, result_seg",
    [
        (0, 2, [1, 0, 2, 3]),
        (0, 3, [2, 3, 0, 1]),
        (1, 2, [0, 1, 2, 3]),
        (1, 3, [0, 1, 3, 2]),
    ],
)
def test_process_two_qubit_gate_combine_segments_cycle2(qb0, qb1, result_seg):
    # Not long enough segment for cyclic
    mapper = lm.LinearMapper(num_qubits=5, cyclic=True)
    segments = [[0, 1], [2, 3]]
    active_qubits = set([0, 1, 2, 3, 4])
    neighbour_ids = {0: set([1]), 1: set([0]), 2: set([3]), 3: set([2])}
    mapper._process_two_qubit_gate(
        num_qubits=mapper.num_qubits,
        cyclic=mapper.cyclic,
        qubit0=qb0,
        qubit1=qb1,
        active_qubits=active_qubits,
        segments=segments,
        neighbour_ids=neighbour_ids,
    )
    assert segments == [result_seg] or segments == [reversed(result_seg)]
    assert qb1 in neighbour_ids[qb0]
    assert qb0 in neighbour_ids[qb1]
    assert result_seg[0] not in neighbour_ids[result_seg[-1]]
    assert result_seg[-1] not in neighbour_ids[result_seg[0]]


@pytest.mark.parametrize(
    "segments, current_chain, correct_chain, allocated_qubits",
    [
        ([[0, 2, 4]], [0, 1, 2, 3, 4], [0, 2, 4, 3, 1], [0, 1, 2, 3, 4]),
        ([[0, 2, 4]], [0, 1, 2, 3, 4], [0, 2, 4, 3, None], [0, 2, 3, 4]),
        ([[1, 2], [3, 0]], [0, 1, 2, 3, 4], [None, 1, 2, 3, 0], [0, 1, 2, 3]),
        ([[1, 2], [3, 0]], [0, 1, 2, 3, 4], [1, 2, 3, 0, 4], [0, 1, 2, 3, 4]),
    ],
)
def test_return_new_mapping_from_segments(segments, current_chain, correct_chain, allocated_qubits):
    mapper = lm.LinearMapper(num_qubits=5, cyclic=False)
    current_mapping = dict()
    for pos, logical_id in enumerate(current_chain):
        current_mapping[logical_id] = pos
    mapper.current_mapping = current_mapping
    new_mapping = mapper._return_new_mapping_from_segments(
        num_qubits=mapper.num_qubits,
        segments=segments,
        allocated_qubits=allocated_qubits,
        current_mapping=mapper.current_mapping,
    )
    correct_mapping = dict()
    for pos, logical_id in enumerate(correct_chain):
        if logical_id is not None:
            correct_mapping[logical_id] = pos
    assert correct_mapping == new_mapping


@pytest.mark.parametrize(
    "old_chain, new_chain",
    [
        ([0, 1, 2, 3, 4], [4, 3, 2, 1, 0]),
        ([2, 0, 14, 44, 12], [14, 12, 44, 0, 2]),
        ([2, None, 14, 44, 12], [14, 1, 44, 0, 2]),
        ([2, None, 14, 44, 12], [14, None, 44, 0, 2]),
    ],
)
def test_odd_even_transposition_sort_swaps(old_chain, new_chain):
    mapper = lm.LinearMapper(num_qubits=5, cyclic=False)
    old_map = dict()
    new_map = dict()
    for pos, logical_id in enumerate(old_chain):
        if logical_id is not None:
            old_map[logical_id] = pos
    for pos, logical_id in enumerate(new_chain):
        if logical_id is not None:
            new_map[logical_id] = pos
    swaps = mapper._odd_even_transposition_sort_swaps(old_map, new_map)
    sorted_chain = deepcopy(old_chain)
    # Remove all ids which are not in new_chain by None
    for i in range(len(sorted_chain)):
        if sorted_chain[i] not in new_chain:
            sorted_chain[i] = None
    for i, j in swaps:
        tmp = sorted_chain[i]
        sorted_chain[i] = sorted_chain[j]
        sorted_chain[j] = tmp
    assert len(sorted_chain) == len(new_chain)
    for i in range(len(sorted_chain)):
        if sorted_chain[i] is not None:
            assert sorted_chain[i] == new_chain[i]


def test_send_possible_commands_allocate():
    mapper = lm.LinearMapper(num_qubits=4, cyclic=False)
    backend = DummyEngine(save_commands=True)
    backend.is_last_engine = True
    mapper.next_engine = backend
    qb0 = WeakQubitRef(engine=None, idx=0)
    cmd0 = Command(engine=None, gate=Allocate, qubits=([qb0],), controls=[], tags=[])
    mapper._stored_commands = [cmd0]
    mapper._currently_allocated_ids = set([10])
    # not in mapping:
    mapper.current_mapping = dict()
    assert len(backend.received_commands) == 0
    mapper._send_possible_commands()
    assert len(backend.received_commands) == 0
    assert mapper._stored_commands == [cmd0]
    # in mapping:
    mapper.current_mapping = {0: 3}
    mapper._send_possible_commands()
    assert len(mapper._stored_commands) == 0
    assert len(backend.received_commands) == 1
    assert backend.received_commands[0].gate == Allocate
    assert backend.received_commands[0].qubits[0][0].id == 3
    assert backend.received_commands[0].tags == [LogicalQubitIDTag(0)]
    assert mapper._currently_allocated_ids == set([10, 0])


def test_send_possible_commands_deallocate():
    mapper = lm.LinearMapper(num_qubits=4, cyclic=False)
    backend = DummyEngine(save_commands=True)
    backend.is_last_engine = True
    mapper.next_engine = backend
    qb0 = WeakQubitRef(engine=None, idx=0)
    cmd0 = Command(engine=None, gate=Deallocate, qubits=([qb0],), controls=[], tags=[])
    mapper._stored_commands = [cmd0]
    mapper.current_mapping = dict()
    mapper._currently_allocated_ids = set([10])
    # not yet allocated:
    mapper._send_possible_commands()
    assert len(backend.received_commands) == 0
    assert mapper._stored_commands == [cmd0]
    # allocated:
    mapper.current_mapping = {0: 3}
    mapper._currently_allocated_ids.add(0)
    mapper._send_possible_commands()
    assert len(backend.received_commands) == 1
    assert backend.received_commands[0].gate == Deallocate
    assert backend.received_commands[0].qubits[0][0].id == 3
    assert backend.received_commands[0].tags == [LogicalQubitIDTag(0)]
    assert len(mapper._stored_commands) == 0
    assert mapper.current_mapping == dict()
    assert mapper._currently_allocated_ids == set([10])


def test_send_possible_commands_keep_remaining_gates():
    mapper = lm.LinearMapper(num_qubits=4, cyclic=False)
    backend = DummyEngine(save_commands=True)
    backend.is_last_engine = True
    mapper.next_engine = backend
    qb0 = WeakQubitRef(engine=None, idx=0)
    qb1 = WeakQubitRef(engine=None, idx=1)
    cmd0 = Command(engine=None, gate=Allocate, qubits=([qb0],), controls=[], tags=[])
    cmd1 = Command(engine=None, gate=Deallocate, qubits=([qb0],), controls=[], tags=[])
    cmd2 = Command(engine=None, gate=Allocate, qubits=([qb1],), controls=[], tags=[])

    mapper._stored_commands = [cmd0, cmd1, cmd2]
    mapper.current_mapping = {0: 0}
    mapper._send_possible_commands()
    assert mapper._stored_commands == [cmd2]


def test_send_possible_commands_not_cyclic():
    mapper = lm.LinearMapper(num_qubits=4, cyclic=False)
    backend = DummyEngine(save_commands=True)
    backend.is_last_engine = True
    mapper.next_engine = backend
    qb0 = WeakQubitRef(engine=None, idx=0)
    qb1 = WeakQubitRef(engine=None, idx=1)
    qb2 = WeakQubitRef(engine=None, idx=2)
    qb3 = WeakQubitRef(engine=None, idx=3)
    mapper._currently_allocated_ids = set([0, 1, 2, 3])
    cmd0 = Command(None, CNOT, qubits=([qb0],), controls=[qb2])
    cmd1 = Command(None, CNOT, qubits=([qb1],), controls=[qb2])
    cmd2 = Command(None, CNOT, qubits=([qb1],), controls=[qb3])
    cmd3 = Command(None, X, qubits=([qb0],), controls=[])
    mapper._stored_commands = [cmd0, cmd1, cmd2, cmd3]
    # Following chain 0 <-> 2 <-> 3 <-> 1
    mapper.current_mapping = {0: 0, 2: 1, 3: 2, 1: 3}
    mapper._send_possible_commands()
    assert len(backend.received_commands) == 2
    assert backend.received_commands[0] == Command(None, CNOT, qubits=([qb0],), controls=[qb1])
    assert backend.received_commands[1] == Command(None, X, qubits=([qb0],))
    # Following chain 0 <-> 2 <-> 1 <-> 3
    mapper.current_mapping = {0: 0, 2: 1, 3: 3, 1: 2}
    mapper._send_possible_commands()
    assert len(backend.received_commands) == 4
    assert len(mapper._stored_commands) == 0


def test_send_possible_commands_cyclic():
    mapper = lm.LinearMapper(num_qubits=4, cyclic=True)
    backend = DummyEngine(save_commands=True)
    backend.is_last_engine = True
    mapper.next_engine = backend
    qb0 = WeakQubitRef(engine=None, idx=0)
    qb1 = WeakQubitRef(engine=None, idx=1)
    qb2 = WeakQubitRef(engine=None, idx=2)
    qb3 = WeakQubitRef(engine=None, idx=3)
    mapper._currently_allocated_ids = set([0, 1, 2, 3])
    cmd0 = Command(None, CNOT, qubits=([qb0],), controls=[qb1])
    cmd1 = Command(None, CNOT, qubits=([qb1],), controls=[qb2])
    cmd2 = Command(None, CNOT, qubits=([qb1],), controls=[qb3])
    cmd3 = Command(None, X, qubits=([qb0],), controls=[])
    mapper._stored_commands = [cmd0, cmd1, cmd2, cmd3]
    # Following chain 0 <-> 2 <-> 3 <-> 1
    mapper.current_mapping = {0: 0, 2: 1, 3: 2, 1: 3}
    mapper._send_possible_commands()
    assert len(backend.received_commands) == 2
    assert backend.received_commands[0] == Command(None, CNOT, qubits=([qb0],), controls=[qb3])
    assert backend.received_commands[1] == Command(None, X, qubits=([qb0],))
    # Following chain 0 <-> 2 <-> 1 <-> 3
    mapper.current_mapping = {0: 0, 2: 1, 3: 3, 1: 2}
    mapper._send_possible_commands()
    assert len(backend.received_commands) == 4
    assert len(mapper._stored_commands) == 0


def test_run_and_receive():
    mapper = lm.LinearMapper(num_qubits=3, cyclic=False)
    backend = DummyEngine(save_commands=True)
    backend.is_last_engine = True
    mapper.next_engine = backend
    qb0 = WeakQubitRef(engine=None, idx=0)
    qb1 = WeakQubitRef(engine=None, idx=1)
    qb2 = WeakQubitRef(engine=None, idx=2)
    cmd0 = Command(engine=None, gate=Allocate, qubits=([qb0],))
    cmd1 = Command(engine=None, gate=Allocate, qubits=([qb1],))
    cmd2 = Command(engine=None, gate=Allocate, qubits=([qb2],))
    cmd3 = Command(None, X, qubits=([qb0],), controls=[qb1])
    cmd4 = Command(None, X, qubits=([qb1],), controls=[qb2])
    cmd5 = Command(engine=None, gate=Deallocate, qubits=([qb1],))
    mapper.receive([cmd0, cmd1, cmd2, cmd3, cmd4, cmd5])
    assert mapper._stored_commands == [cmd0, cmd1, cmd2, cmd3, cmd4, cmd5]
    qb3 = WeakQubitRef(engine=None, idx=-1)
    cmd_flush = Command(engine=None, gate=FlushGate(), qubits=([qb3],))
    mapper.receive([cmd_flush])
    assert mapper._stored_commands == []
    assert len(backend.received_commands) == 7
    assert mapper._currently_allocated_ids == set([0, 2])
    assert mapper.current_mapping == {0: 2, 2: 0} or mapper.current_mapping == {
        0: 0,
        2: 2,
    }
    cmd6 = Command(None, X, qubits=([qb0],), controls=[qb2])
    mapper.storage = 1
    mapper.receive([cmd6])
    assert mapper._currently_allocated_ids == set([0, 2])
    assert mapper._stored_commands == []
    assert len(mapper.current_mapping) == 2
    assert 0 in mapper.current_mapping
    assert 2 in mapper.current_mapping
    assert len(backend.received_commands) == 11
    for cmd in backend.received_commands:
        print(cmd)
    assert backend.received_commands[-1] == Command(
        None,
        X,
        qubits=([WeakQubitRef(engine=None, idx=mapper.current_mapping[qb0.id])],),
        controls=[WeakQubitRef(engine=None, idx=mapper.current_mapping[qb2.id])],
    )
    assert mapper.num_mappings == 1


def test_run_infinite_loop_detection():
    mapper = lm.LinearMapper(num_qubits=1, cyclic=False)
    backend = DummyEngine(save_commands=True)
    backend.is_last_engine = True
    mapper.next_engine = backend
    qb0 = WeakQubitRef(engine=None, idx=0)
    qb1 = WeakQubitRef(engine=None, idx=1)
    cmd0 = Command(engine=None, gate=Allocate, qubits=([qb0],))
    cmd1 = Command(engine=None, gate=Allocate, qubits=([qb1],))
    cmd2 = Command(None, X, qubits=([qb0],), controls=[qb1])
    qb2 = WeakQubitRef(engine=None, idx=-1)
    cmd_flush = Command(engine=None, gate=FlushGate(), qubits=([qb2],))
    with pytest.raises(RuntimeError):
        mapper.receive([cmd0, cmd1, cmd2, cmd_flush])


def test_logical_id_tags_allocate_and_deallocate():
    mapper = lm.LinearMapper(num_qubits=4, cyclic=False)
    backend = DummyEngine(save_commands=True)
    backend.is_last_engine = True
    mapper.next_engine = backend
    qb0 = WeakQubitRef(engine=None, idx=0)
    qb1 = WeakQubitRef(engine=None, idx=1)
    cmd0 = Command(engine=None, gate=Allocate, qubits=([qb0],))
    cmd1 = Command(engine=None, gate=Allocate, qubits=([qb1],))
    cmd2 = Command(None, X, qubits=([qb0],), controls=[qb1])
    cmd3 = Command(engine=None, gate=Deallocate, qubits=([qb0],))
    cmd4 = Command(engine=None, gate=Deallocate, qubits=([qb1],))
    mapper.current_mapping = {0: 0, 1: 3}
    qb_flush = WeakQubitRef(engine=None, idx=-1)
    cmd_flush = Command(engine=None, gate=FlushGate(), qubits=([qb_flush],))
    mapper.receive([cmd0, cmd1, cmd2, cmd_flush])
    assert backend.received_commands[0].gate == Allocate
    assert backend.received_commands[0].qubits[0][0].id == 0
    assert backend.received_commands[0].tags == [LogicalQubitIDTag(0)]
    assert backend.received_commands[1].gate == Allocate
    assert backend.received_commands[1].qubits[0][0].id == 3
    assert backend.received_commands[1].tags == [LogicalQubitIDTag(1)]
    for cmd in backend.received_commands[2:]:
        if cmd.gate == Allocate:
            assert cmd.tags == []
        elif cmd.gate == Deallocate:
            assert cmd.tags == []
    mapped_id_for_0 = mapper.current_mapping[0]
    mapped_id_for_1 = mapper.current_mapping[1]
    mapper.receive([cmd3, cmd4, cmd_flush])
    assert backend.received_commands[-3].gate == Deallocate
    assert backend.received_commands[-3].qubits[0][0].id == mapped_id_for_0
    assert backend.received_commands[-3].tags == [LogicalQubitIDTag(0)]
    assert backend.received_commands[-2].gate == Deallocate
    assert backend.received_commands[-2].qubits[0][0].id == mapped_id_for_1
    assert backend.received_commands[-2].tags == [LogicalQubitIDTag(1)]


def test_send_possible_cmds_before_new_mapping():
    mapper = lm.LinearMapper(num_qubits=3, cyclic=False)
    backend = DummyEngine(save_commands=True)
    backend.is_last_engine = True
    mapper.next_engine = backend

    def dont_call_mapping():
        raise Exception

    mapper._return_new_mapping = dont_call_mapping
    mapper.current_mapping = {0: 1}
    qb0 = WeakQubitRef(engine=None, idx=0)
    cmd0 = Command(engine=None, gate=Allocate, qubits=([qb0],))
    qb2 = WeakQubitRef(engine=None, idx=-1)
    cmd_flush = Command(engine=None, gate=FlushGate(), qubits=([qb2],))
    mapper.receive([cmd0, cmd_flush])


def test_correct_stats():
    # Should test stats for twice same mapping but depends on heuristic
    mapper = lm.LinearMapper(num_qubits=3, cyclic=False)
    backend = DummyEngine(save_commands=True)
    backend.is_last_engine = True
    mapper.next_engine = backend
    qb0 = WeakQubitRef(engine=None, idx=0)
    qb1 = WeakQubitRef(engine=None, idx=1)
    qb2 = WeakQubitRef(engine=None, idx=2)
    cmd0 = Command(engine=None, gate=Allocate, qubits=([qb0],))
    cmd1 = Command(engine=None, gate=Allocate, qubits=([qb1],))
    cmd2 = Command(engine=None, gate=Allocate, qubits=([qb2],))
    cmd3 = Command(None, X, qubits=([qb0],), controls=[qb1])
    cmd4 = Command(None, X, qubits=([qb1],), controls=[qb2])
    cmd5 = Command(None, X, qubits=([qb0],), controls=[qb2])
    cmd6 = Command(None, X, qubits=([qb2],), controls=[qb1])
    cmd7 = Command(None, X, qubits=([qb0],), controls=[qb1])
    cmd8 = Command(None, X, qubits=([qb1],), controls=[qb2])
    qb_flush = WeakQubitRef(engine=None, idx=-1)
    cmd_flush = Command(engine=None, gate=FlushGate(), qubits=([qb_flush],))
    mapper.receive([cmd0, cmd1, cmd2, cmd3, cmd4, cmd5, cmd6, cmd7, cmd8, cmd_flush])
    assert mapper.num_mappings == 2



# -*- coding: utf-8 -*-
#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Tests for libs.revkit._phase."""

import pytest

from projectq import MainEngine
from projectq.backends import Simulator
from projectq.cengines import DummyEngine
from projectq.ops import All, H, Measure

from projectq.libs.revkit import PhaseOracle

import numpy as np

# run this test only if RevKit Python module can be loaded
revkit = pytest.importorskip('revkit')


def test_phase_majority():
    sim = Simulator()
    main_engine = MainEngine(sim)

    qureg = main_engine.allocate_qureg(3)
    All(H) | qureg
    PhaseOracle(0xE8) | qureg

    main_engine.flush()

    assert np.array_equal(np.sign(sim.cheat()[1]), [1.0, 1.0, 1.0, -1.0, 1.0, -1.0, -1.0, -1.0])
    All(Measure) | qureg


def test_phase_majority_from_python():
    dormouse = pytest.importorskip('dormouse')  # noqa: F841

    def maj(a, b, c):
        return (a and b) or (a and c) or (b and c)  # pragma: no cover

    sim = Simulator()
    main_engine = MainEngine(sim)

    qureg = main_engine.allocate_qureg(3)
    All(H) | qureg
    PhaseOracle(maj) | qureg

    main_engine.flush()

    assert np.array_equal(np.sign(sim.cheat()[1]), [1.0, 1.0, 1.0, -1.0, 1.0, -1.0, -1.0, -1.0])
    All(Measure) | qureg


def test_phase_invalid_function():
    main_engine = MainEngine(backend=DummyEngine(), engine_list=[DummyEngine()])

    qureg = main_engine.allocate_qureg(3)

    with pytest.raises(AttributeError):
        PhaseOracle(-42) | qureg

    with pytest.raises(AttributeError):
        PhaseOracle(0xCAFE) | qureg

    with pytest.raises(RuntimeError):
        PhaseOracle(0x8E, synth=lambda: revkit.esopbs()) | qureg



# -*- coding: utf-8 -*-
#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""
Contains the tools to make an entire section of operations controlled.

Example:
    .. code-block:: python

        with Control(eng, qubit1):
            H | qubit2
            X | qubit3
"""

from projectq.cengines import BasicEngine
from projectq.ops import ClassicalInstructionGate, CtrlAll
from projectq.types import BasicQubit

from ._compute import ComputeTag, UncomputeTag
from ._util import insert_engine, drop_engine_after


def canonical_ctrl_state(ctrl_state, num_qubits):
    """
    Return canonical form for control state

    Args:
        ctrl_state (int,str,CtrlAll): Initial control state representation
        num_qubits (int): number of control qubits

    Returns:
        Canonical form of control state (currently a string composed of '0' and '1')

    Note:
        In case of integer values for `ctrl_state`, the least significant bit applies to the first qubit in the qubit
        register, e.g. if ctrl_state == 2, its binary representation if '10' with the least significan bit being 0.

        This means in particular that the followings are equivalent:

        .. code-block:: python

            canonical_ctrl_state(6, 3) == canonical_ctrl_state(6, '110')
    """
    if not num_qubits:
        return ''

    if isinstance(ctrl_state, CtrlAll):
        if ctrl_state == CtrlAll.One:
            return '1' * num_qubits
        return '0' * num_qubits

    if isinstance(ctrl_state, int):
        # If the user inputs an integer, convert it to binary bit string
        converted_str = '{0:b}'.format(ctrl_state).zfill(num_qubits)[::-1]
        if len(converted_str) != num_qubits:
            raise ValueError(
                'Control state specified as {} ({}) is higher than maximum for {} qubits: {}'.format(
                    ctrl_state, converted_str, num_qubits, 2 ** num_qubits - 1
                )
            )
        return converted_str

    if isinstance(ctrl_state, str):
        # If the user inputs bit string, directly use it
        if len(ctrl_state) != num_qubits:
            raise ValueError(
                'Control state {} has different length than the number of control qubits {}'.format(
                    ctrl_state, num_qubits
                )
            )
        if not set(ctrl_state).issubset({'0', '1'}):
            raise ValueError('Control state {} has string other than 1 and 0'.format(ctrl_state))
        return ctrl_state

    raise TypeError('Input must be a string, an integer or an enum value of class State')


def _has_compute_uncompute_tag(cmd):
    """
    Return True if command cmd has a compute/uncompute tag.

    Args:
        cmd (Command object): a command object.
    """
    for tag in cmd.tags:
        if tag in [UncomputeTag(), ComputeTag()]:
            return True
    return False


class ControlEngine(BasicEngine):
    """
    Adds control qubits to all commands that have no compute / uncompute tags.
    """

    def __init__(self, qubits, ctrl_state=CtrlAll.One):
        """
        Initialize the control engine.

        Args:
            qubits (list of Qubit objects): qubits conditional on which the
                following operations are executed.
        """
        BasicEngine.__init__(self)
        self._qubits = qubits
        self._state = ctrl_state

    def _handle_command(self, cmd):
        if not _has_compute_uncompute_tag(cmd) and not isinstance(cmd.gate, ClassicalInstructionGate):
            cmd.add_control_qubits(self._qubits, self._state)
        self.send([cmd])

    def receive(self, command_list):
        """Forward all commands to the next engine."""
        for cmd in command_list:
            self._handle_command(cmd)


class Control:
    """
    Condition an entire code block on the value of qubits being 1.

    Example:
        .. code-block:: python

            with Control(eng, ctrlqubits):
                do_something(otherqubits)
    """

    def __init__(self, engine, qubits, ctrl_state=CtrlAll.One):
        """
        Enter a controlled section.

        Args:
            engine: Engine which handles the commands (usually MainEngine)
            qubits (list of Qubit objects): Qubits to condition on

        Enter the section using a with-statement:

        .. code-block:: python

            with Control(eng, ctrlqubits):
                ...
        """
        self.engine = engine
        if isinstance(qubits, tuple):
            raise TypeError('Control qubits must be a list, not a tuple!')
        if isinstance(qubits, BasicQubit):
            qubits = [qubits]
        self._qubits = qubits
        self._state = canonical_ctrl_state(ctrl_state, len(self._qubits))

    def __enter__(self):
        if len(self._qubits) > 0:
            engine = ControlEngine(self._qubits, self._state)
            insert_engine(self.engine, engine)

    def __exit__(self, exc_type, exc_value, exc_traceback):
        # remove control handler from engine list (i.e. skip it)
        if len(self._qubits) > 0:
            drop_engine_after(self.engine)


def get_control_count(cmd):
    """
    Return the number of control qubits of the command object cmd
    """
    return len(cmd.control_qubits)


def has_negative_control(cmd):
    """
    Returns whether a command has negatively controlled qubits
    """
    return get_control_count(cmd) > 0 and '0' in cmd.control_state



# -*- coding: utf-8 -*-
#   Copyright 2021 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
# pylint: skip-file

"""Example of a basic entangling operation using an IonQBackend."""

import getpass

import matplotlib.pyplot as plt

import projectq.setups.ionq
from projectq import MainEngine
from projectq.backends import IonQBackend
from projectq.libs.hist import histogram
from projectq.ops import All, Entangle, Measure


def run_entangle(eng, num_qubits=3):
    """
    Runs an entangling operation on the provided compiler engine.

    Args:
        eng (MainEngine): Main compiler engine to use.
        num_qubits (int): Number of qubits to entangle.

    Returns:
        measurement (list<int>): List of measurement outcomes.
    """
    # allocate the quantum register to entangle
    qureg = eng.allocate_qureg(num_qubits)

    # entangle the qureg
    Entangle | qureg

    # measure; should be all-0 or all-1
    All(Measure) | qureg

    # run the circuit
    eng.flush()

    # access the probabilities via the back-end:
    # results = eng.backend.get_probabilities(qureg)
    # for state in results:
    #     print("Measured {} with p = {}.".format(state, results[state]))
    # or plot them directly:
    histogram(eng.backend, qureg)
    plt.show()

    # return one (random) measurement outcome.
    return [int(q) for q in qureg]


if __name__ == '__main__':
    token = None
    device = None
    if token is None:
        token = getpass.getpass(prompt='IonQ apiKey > ')
    if device is None:
        device = input('IonQ device > ')

    # create an IonQBackend
    backend = IonQBackend(
        use_hardware=True,
        token=token,
        num_runs=200,
        verbose=True,
        device=device,
    )
    engine_list = projectq.setups.ionq.get_engine_list(
        token=token,
        device=device,
    )
    engine = MainEngine(backend, engine_list)
    # run the circuit and print the result
    print(run_entangle(engine))



# -*- coding: utf-8 -*-
#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"Tests for projectq.setups.decompositions.carb1qubit2cnotrzandry.py."

import numpy as np
import pytest

from projectq.backends import Simulator
from projectq.cengines import (
    AutoReplacer,
    DecompositionRuleSet,
    DummyEngine,
    InstructionFilter,
    MainEngine,
)
from projectq.meta import Control
from projectq.ops import (
    All,
    BasicGate,
    ClassicalInstructionGate,
    MatrixGate,
    Measure,
    Ph,
    R,
    Rx,
    Ry,
    Rz,
    X,
    XGate,
)
from projectq.setups.decompositions import arb1qubit2rzandry_test as arb1q_t

from . import carb1qubit2cnotrzandry as carb1q


def test_recognize_correct_gates():
    saving_backend = DummyEngine(save_commands=True)
    eng = MainEngine(backend=saving_backend)
    qubit = eng.allocate_qubit()
    ctrl_qubit = eng.allocate_qubit()
    eng.flush()
    with Control(eng, ctrl_qubit):
        Ph(0.1) | qubit
        R(0.2) | qubit
        Rx(0.3) | qubit
        X | qubit
    eng.flush(deallocate_qubits=True)
    # Don't test initial two allocate and flush and trailing deallocate
    # and flush gate.
    for cmd in saving_backend.received_commands[3:-3]:
        assert carb1q._recognize_carb1qubit(cmd)


def test_recognize_incorrect_gates():
    saving_backend = DummyEngine(save_commands=True)
    eng = MainEngine(backend=saving_backend)
    qubit = eng.allocate_qubit()
    ctrl_qubit = eng.allocate_qubit()
    ctrl_qureg = eng.allocate_qureg(2)
    eng.flush()
    with Control(eng, ctrl_qubit):
        # Does not have matrix attribute:
        BasicGate() | qubit
        # Two qubit gate:
        two_qubit_gate = MatrixGate()
        two_qubit_gate.matrix = np.matrix([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]])
        two_qubit_gate | qubit
    with Control(eng, ctrl_qureg):
        # Too many Control qubits:
        Rx(0.3) | qubit
    eng.flush(deallocate_qubits=True)
    for cmd in saving_backend.received_commands:
        assert not carb1q._recognize_carb1qubit(cmd)


def _decomp_gates(eng, cmd):
    g = cmd.gate
    if isinstance(g, ClassicalInstructionGate):
        return True
    if len(cmd.control_qubits) == 0:
        if isinstance(cmd.gate, Ry) or isinstance(cmd.gate, Rz) or isinstance(cmd.gate, Ph):
            return True
    if len(cmd.control_qubits) == 1:
        if isinstance(cmd.gate, XGate) or isinstance(cmd.gate, Ph):
            return True
    return False


# yapf: disable
@pytest.mark.parametrize("gate_matrix", [[[1, 0], [0, -1]],
                                         [[0, -1j], [1j, 0]]])
def test_recognize_v(gate_matrix):
    assert carb1q._recognize_v(gate_matrix)
# yapf: enable


@pytest.mark.parametrize("gate_matrix", arb1q_t.create_test_matrices())
def test_decomposition(gate_matrix):
    # Create single qubit gate with gate_matrix
    test_gate = MatrixGate()
    test_gate.matrix = np.matrix(gate_matrix)

    for basis_state in ([1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]):
        correct_dummy_eng = DummyEngine(save_commands=True)
        correct_eng = MainEngine(backend=Simulator(), engine_list=[correct_dummy_eng])

        rule_set = DecompositionRuleSet(modules=[carb1q])
        test_dummy_eng = DummyEngine(save_commands=True)
        test_eng = MainEngine(
            backend=Simulator(),
            engine_list=[
                AutoReplacer(rule_set),
                InstructionFilter(_decomp_gates),
                test_dummy_eng,
            ],
        )
        test_sim = test_eng.backend
        correct_sim = correct_eng.backend

        correct_qb = correct_eng.allocate_qubit()
        correct_ctrl_qb = correct_eng.allocate_qubit()
        correct_eng.flush()
        test_qb = test_eng.allocate_qubit()
        test_ctrl_qb = test_eng.allocate_qubit()
        test_eng.flush()

        correct_sim.set_wavefunction(basis_state, correct_qb + correct_ctrl_qb)
        test_sim.set_wavefunction(basis_state, test_qb + test_ctrl_qb)

        with Control(test_eng, test_ctrl_qb):
            test_gate | test_qb
        with Control(correct_eng, correct_ctrl_qb):
            test_gate | correct_qb

        test_eng.flush()
        correct_eng.flush()

        assert correct_dummy_eng.received_commands[3].gate == test_gate
        assert test_dummy_eng.received_commands[3].gate != test_gate

        for fstate in ['00', '01', '10', '11']:
            test = test_sim.get_amplitude(fstate, test_qb + test_ctrl_qb)
            correct = correct_sim.get_amplitude(fstate, correct_qb + correct_ctrl_qb)
            assert correct == pytest.approx(test, rel=1e-12, abs=1e-12)

        All(Measure) | test_qb + test_ctrl_qb
        All(Measure) | correct_qb + correct_ctrl_qb
        test_eng.flush(deallocate_qubits=True)
        correct_eng.flush(deallocate_qubits=True)



# -*- coding: utf-8 -*-
# pylint: skip-file

from projectq.cengines import MainEngine
from projectq.ops import All, H, X, Measure
from projectq.meta import Compute, Uncompute
from projectq.libs.revkit import PhaseOracle


# phase function
def f(a, b, c, d):
    return (a and b) ^ (c and d)


eng = MainEngine()
x1, x2, x3, x4 = qubits = eng.allocate_qureg(4)

with Compute(eng):
    All(H) | qubits
    X | x1
PhaseOracle(f) | qubits
Uncompute(eng)

PhaseOracle(f) | qubits
All(H) | qubits
All(Measure) | qubits

eng.flush()

print("Shift is {}".format(8 * int(x4) + 4 * int(x3) + 2 * int(x2) + int(x1)))



# -*- coding: utf-8 -*-
#   Copyright 2021 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

""" HTTP Client for the IonQ API. """

import getpass
import json
import signal
import time

import requests
from requests import Session
from requests.compat import urljoin

from ._ionq_exc import (
    DeviceOfflineError,
    DeviceTooSmall,
    JobSubmissionError,
    RequestTimeoutError,
)

_API_URL = 'https://api.ionq.co/v0.1/jobs/'


class IonQ(Session):
    """A requests.Session based HTTP client for the IonQ API."""

    def __init__(self, verbose=False):
        super().__init__()
        self.backends = dict()
        self.timeout = 5.0
        self.token = None
        self._verbose = verbose

    def update_devices_list(self):
        """Update the list of devices this backend can support."""
        self.backends = {
            'ionq_simulator': {
                'nq': 29,
                'target': 'simulator',
            },
            'ionq_qpu': {
                'nq': 11,
                'target': 'qpu',
            },
        }
        if self._verbose:  # pragma: no cover
            print('- List of IonQ devices available:')
            print(self.backends)

    def is_online(self, device):
        """Check if a given device is online.

        Args:
            device (str): An IonQ device name.

        Returns:
            bool: True if device is online, else False.
        """
        return device in self.backends

    def can_run_experiment(self, info, device):
        """
        Determine whether or not the desired device has enough allocatable
        qubits to run something.

        This returns a three-element tuple with whether or not the experiment
        can be run, the max number of qubits possible, and the number of qubits
        needed to run this experiment.

        Args:
            info (dict): A dict containing number of shots, qubits, and
                a circuit.
            device (str): An IonQ device name.
        Returns:
            tuple(bool, int, int): Whether the operation can be run, max
                number of qubits the device supports, and number of qubits
                required for the experiment.
        """
        nb_qubit_max = self.backends[device]['nq']
        nb_qubit_needed = info['nq']
        return nb_qubit_needed <= nb_qubit_max, nb_qubit_max, nb_qubit_needed

    def authenticate(self, token=None):
        """Set an Authorization header for this session.

        If no token is provided, an prompt will appear to ask for one.

        Args:
            token (str): IonQ user API token.
        """
        if token is None:
            token = getpass.getpass(prompt='IonQ apiKey > ')
        if not token:
            raise RuntimeError('An authentication token is required!')
        self.headers.update({'Authorization': 'apiKey {}'.format(token)})
        self.token = token

    def run(self, info, device):
        """Run a circuit from ``info`` on the specified ``device``.

        Args:
            info (dict): A dict containing number of shots, qubits, and
                a circuit.
            device (str): An IonQ device name.

        Raises:
            JobSubmissionError: If the job creation response from IonQ's API
                had a failure result.

        Returns:
            str: The ID of a newly submitted Job.
        """
        argument = {
            'target': self.backends[device]['target'],
            'metadata': {
                'sdk': 'ProjectQ',
                'meas_qubit_ids': json.dumps(info['meas_qubit_ids']),
            },
            'shots': info['shots'],
            'registers': {'meas_mapped': info['meas_mapped']},
            'lang': 'json',
            'body': {
                'qubits': info['nq'],
                'circuit': info['circuit'],
            },
        }

        # _API_URL[:-1] strips the trailing slash.
        # TODO: Add comprehensive error parsing for non-200 responses.
        req = super().post(_API_URL[:-1], json=argument)
        req.raise_for_status()

        # Process the response.
        r_json = req.json()
        status = r_json['status']

        # Return the job id.
        if status == 'ready':
            return r_json['id']

        # Otherwise, extract any provided failure info and raise an exception.
        failure = r_json.get('failure') or {
            'code': 'UnknownError',
            'error': 'An unknown error occurred!',
        }
        raise JobSubmissionError(
            "{}: {} (status={})".format(
                failure['code'],
                failure['error'],
                status,
            )
        )

    def get_result(self, device, execution_id, num_retries=3000, interval=1):
        """Given a backend and ID, fetch the results for this job's execution.

        The return dictionary should have at least:

            * ``nq`` (int): Number of qubits for this job.
            * ``output_probs`` (dict): Map of integer states to probability values.

        Args:
            device (str): The device used to run this job.
            execution_id (str): An IonQ Job ID.
            num_retries (int, optional): Number of times to retry the fetch
                before raising a timeout error. Defaults to 3000.
            interval (int, optional): Number of seconds to wait between retries.
                Defaults to 1.

        Raises:
            Exception: If the process receives a kill signal before completion.
            Exception: If the job is in an unknown processing state.
            DeviceOfflineError: If the provided device is not online.
            RequestTimeoutError: If we were unable to retrieve the job results
                after ``num_retries`` attempts.

        Returns:
            dict: A dict of job data for an engine to consume.
        """

        if self._verbose:  # pragma: no cover
            print("Waiting for results. [Job ID: {}]".format(execution_id))

        original_sigint_handler = signal.getsignal(signal.SIGINT)

        def _handle_sigint_during_get_result(*_):  # pragma: no cover
            raise Exception("Interrupted. The ID of your submitted job is {}.".format(execution_id))

        signal.signal(signal.SIGINT, _handle_sigint_during_get_result)

        try:
            for retries in range(num_retries):
                req = super().get(urljoin(_API_URL, execution_id))
                req.raise_for_status()
                r_json = req.json()
                status = r_json['status']

                # Check if job is completed.
                if status == 'completed':
                    meas_mapped = r_json['registers']['meas_mapped']
                    meas_qubit_ids = json.loads(r_json['metadata']['meas_qubit_ids'])
                    output_probs = r_json['data']['registers']['meas_mapped']
                    return {
                        'nq': r_json['qubits'],
                        'output_probs': output_probs,
                        'meas_mapped': meas_mapped,
                        'meas_qubit_ids': meas_qubit_ids,
                    }

                # Otherwise, make sure it is in a known healthy state.
                if status not in ('ready', 'running', 'submitted'):
                    # TODO: Add comprehensive API error processing here.
                    raise Exception("Error while running the code: {}.".format(status))

                # Sleep, then check availability before trying again.
                time.sleep(interval)
                if self.is_online(device) and retries % 60 == 0:
                    self.update_devices_list()
                    if not self.is_online(device):  # pragma: no cover
                        raise DeviceOfflineError(
                            "Device went offline. The ID of " "your submitted job is {}.".format(execution_id)
                        )
        finally:
            if original_sigint_handler is not None:
                signal.signal(signal.SIGINT, original_sigint_handler)

        raise RequestTimeoutError("Timeout. The ID of your submitted job is {}.".format(execution_id))


def show_devices(verbose=False):
    """Show the currently available device list for the IonQ provider.

    Args:
        verbose (bool): If True, additional information is printed

    Returns:
        list: list of available devices and their properties.
    """
    ionq_session = IonQ(verbose=verbose)
    ionq_session.update_devices_list()
    return ionq_session.backends


def retrieve(
    device,
    token,
    jobid,
    num_retries=3000,
    interval=1,
    verbose=False,
):  # pylint: disable=too-many-arguments
    """Retrieve an already submitted IonQ job.

    Args:
        device (str): The name of an IonQ device.
        token (str): An IonQ API token.
        jobid (str): An IonQ Job ID.
        num_retries (int, optional): Number of times to retry while the job is
            not finished. Defaults to 3000.
        interval (int, optional): Sleep interval between retries, in seconds.
            Defaults to 1.
        verbose (bool, optional): Whether to print verbose output.
            Defaults to False.

    Returns:
        dict: A dict with job submission results.
    """
    ionq_session = IonQ(verbose=verbose)
    ionq_session.authenticate(token)
    ionq_session.update_devices_list()
    res = ionq_session.get_result(
        device,
        jobid,
        num_retries=num_retries,
        interval=interval,
    )
    return res


def send(
    info,
    device='ionq_simulator',
    token=None,
    num_retries=100,
    interval=1,
    verbose=False,
):  # pylint: disable=too-many-arguments,too-many-locals
    """Submit a job to the IonQ API.

    The ``info`` dict should have at least the following keys::

        * nq (int): Number of qubits this job will need.
        * shots (dict): The number of shots to use for this job.
        * meas_mapped (list): A list of qubits to measure.
        * circuit (list): A list of JSON-serializable IonQ gate representations.

    Args:
        info (dict): A dictionary with
        device (str, optional): The IonQ device to run this on. Defaults to 'ionq_simulator'.
        token (str, optional): An IonQ API token. Defaults to None.
        num_retries (int, optional): Number of times to retry while the job is
            not finished. Defaults to 100.
        interval (int, optional): Sleep interval between retries, in seconds.
            Defaults to 1.
        verbose (bool, optional): Whether to print verbose output.
            Defaults to False.

    Raises:
        DeviceOfflineError: If the desired device is not available for job
            processing.
        DeviceTooSmall: If the job has a higher qubit requirement than the
            device supports.

    Returns:
        dict: An intermediate dict representation of an IonQ job result.
    """
    try:
        ionq_session = IonQ(verbose=verbose)

        if verbose:  # pragma: no cover
            print("- Authenticating...")
        if verbose and token is not None:  # pragma: no cover
            print('user API token: ' + token)
        ionq_session.authenticate(token)

        # check if the device is online
        ionq_session.update_devices_list()
        online = ionq_session.is_online(device)

        # useless for the moment
        if not online:  # pragma: no cover
            print("The device is offline (for maintenance?). Use the " "simulator instead or try again later.")
            raise DeviceOfflineError("Device is offline.")

        # check if the device has enough qubit to run the code
        runnable, qmax, qneeded = ionq_session.can_run_experiment(info, device)
        if not runnable:
            print(
                "The device is too small ({} qubits available) for the code "
                "requested({} qubits needed). Try to look for another device "
                "with more qubits".format(qmax, qneeded)
            )
            raise DeviceTooSmall("Device is too small.")
        if verbose:  # pragma: no cover
            print("- Running code: {}".format(info))
        execution_id = ionq_session.run(info, device)
        if verbose:  # pragma: no cover
            print("- Waiting for results...")
        res = ionq_session.get_result(
            device,
            execution_id,
            num_retries=num_retries,
            interval=interval,
        )
        if verbose:  # pragma: no cover
            print("- Done.")
        return res
    except requests.exceptions.HTTPError as err:
        # Re-raise auth errors, as literally nothing else will work.
        if err.response is not None:
            status_code = err.response.status_code
            if status_code in (401, 403):
                raise err

            # Try to parse client errors
            if status_code == 400:
                err_json = err.response.json()
                raise JobSubmissionError(
                    '{}: {}'.format(
                        err_json['error'],
                        err_json['message'],
                    )
                ) from err

        # Else, just print:
        print("- There was an error running your code:")
        print(err)
    except requests.exceptions.RequestException as err:
        print("- Looks like something is wrong with server:")
        print(err)
    return None


__all__ = [
    'send',
    'retrieve',
    'show_devices',
    'IonQ',
]



# -*- coding: utf-8 -*-
#   Copyright 2021 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
Defines a setup allowing to compile code for IonQ trapped ion devices:
->The 11 qubit device
->The 29 qubits simulator
"""
from projectq.backends._ionq._ionq_exc import DeviceOfflineError
from projectq.backends._ionq._ionq_http_client import show_devices
from projectq.backends._ionq._ionq_mapper import BoundedQubitMapper
from projectq.ops import (
    Barrier,
    H,
    Rx,
    Rxx,
    Ry,
    Ryy,
    Rz,
    Rzz,
    S,
    Sdag,
    SqrtX,
    Swap,
    T,
    Tdag,
    X,
    Y,
    Z,
)
from projectq.setups import restrictedgateset


def get_engine_list(token=None, device=None):
    """
    Return the default list of compiler engine for the IonQ platform
    """
    devices = show_devices(token)
    if not device or device not in devices:
        raise DeviceOfflineError("Error checking engine list: no '{}' devices available".format(device))

    #
    # Qubit mapper
    #
    mapper = BoundedQubitMapper(devices[device]['nq'])

    #
    # Basis Gates
    #

    # Declare the basis gateset for the IonQ's API.
    engine_list = restrictedgateset.get_engine_list(
        one_qubit_gates=(X, Y, Z, Rx, Ry, Rz, H, S, Sdag, T, Tdag, SqrtX),
        two_qubit_gates=(Swap, Rxx, Ryy, Rzz),
        other_gates=(Barrier,),
    )
    return engine_list + [mapper]


__all__ = ['get_engine_list']



# -*- coding: utf-8 -*-
#   Copyright 2021 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Tests for projectq.setup.ionq."""

import pytest

from projectq.backends._ionq._ionq_exc import DeviceOfflineError
from projectq.backends._ionq._ionq_mapper import BoundedQubitMapper


def test_basic_ionq_mapper(monkeypatch):
    import projectq.setups.ionq

    def mock_show_devices(*args, **kwargs):
        return {'dummy': {'nq': 3, 'target': 'dummy'}}

    monkeypatch.setattr(projectq.setups.ionq, 'show_devices', mock_show_devices)
    engine_list = projectq.setups.ionq.get_engine_list(device='dummy')
    assert len(engine_list) > 1
    mapper = engine_list[-1]
    assert isinstance(mapper, BoundedQubitMapper)
    # to match nq in the backend
    assert mapper.max_qubits == 3


def test_ionq_errors(monkeypatch):
    import projectq.setups.ionq

    def mock_show_devices(*args, **kwargs):
        return {'dummy': {'nq': 3, 'target': 'dummy'}}

    monkeypatch.setattr(projectq.setups.ionq, 'show_devices', mock_show_devices)

    with pytest.raises(DeviceOfflineError):
        projectq.setups.ionq.get_engine_list(device='simulator')



# -*- coding: utf-8 -*-
#   Copyright 2018 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"Tests for projectq.setups.decompositions.cnot2rxx.py."

import pytest

from projectq import MainEngine
from projectq.backends import Simulator
from projectq.cengines import (
    AutoReplacer,
    DecompositionRuleSet,
    DummyEngine,
    InstructionFilter,
)
from projectq.meta import Control
from projectq.ops import All, CNOT, CZ, Measure, X, Z

from . import cnot2rxx


def test_recognize_correct_gates():
    """Test that recognize_cnot recognizes cnot gates."""
    saving_backend = DummyEngine(save_commands=True)
    eng = MainEngine(backend=saving_backend)
    qubit1 = eng.allocate_qubit()
    qubit2 = eng.allocate_qubit()
    qubit3 = eng.allocate_qubit()
    eng.flush()
    # Create a control function in 3 different ways
    CZ | (qubit1, qubit2)
    with Control(eng, qubit2):
        Z | qubit1
        X | qubit1
    with Control(eng, qubit2 + qubit3):
        Z | qubit1
    eng.flush()
    eng.flush(deallocate_qubits=True)
    for cmd in saving_backend.received_commands[4:7]:
        assert cnot2rxx._recognize_cnot2(cmd)
    for cmd in saving_backend.received_commands[7:9]:
        assert not cnot2rxx._recognize_cnot2(cmd)


def _decomp_gates(eng, cmd):
    """Test that the cmd.gate is a gate of class X"""
    if len(cmd.control_qubits) == 1 and isinstance(cmd.gate, X.__class__):
        return False
    return True


# ------------test_decomposition function-------------#
# Creates two engines, correct_eng and test_eng.
# correct_eng implements CNOT gate.
# test_eng implements the decomposition of the CNOT gate.
# correct_qb and test_qb represent results of these two engines, respectively.
#
# The decomposition in this case only produces the same state as CNOT up to a
# global phase.
# test_vector and correct_vector represent the final wave states of correct_qb
# and test_qb.
#
# The dot product of correct_vector and test_vector should have absolute value
# 1, if the two vectors are the same up to a global phase.


def test_decomposition():
    """Test that this decomposition of CNOT produces correct amplitudes

    Function tests each DecompositionRule in
    cnot2rxx.all_defined_decomposition_rules
    """
    decomposition_rule_list = cnot2rxx.all_defined_decomposition_rules
    for rule in decomposition_rule_list:
        for basis_state_index in range(0, 4):
            basis_state = [0] * 4
            basis_state[basis_state_index] = 1.0
            correct_dummy_eng = DummyEngine(save_commands=True)
            correct_eng = MainEngine(backend=Simulator(), engine_list=[correct_dummy_eng])
            rule_set = DecompositionRuleSet(rules=[rule])
            test_dummy_eng = DummyEngine(save_commands=True)
            test_eng = MainEngine(
                backend=Simulator(),
                engine_list=[
                    AutoReplacer(rule_set),
                    InstructionFilter(_decomp_gates),
                    test_dummy_eng,
                ],
            )
            test_sim = test_eng.backend
            correct_sim = correct_eng.backend
            correct_qb = correct_eng.allocate_qubit()
            correct_ctrl_qb = correct_eng.allocate_qubit()
            correct_eng.flush()
            test_qb = test_eng.allocate_qubit()
            test_ctrl_qb = test_eng.allocate_qubit()
            test_eng.flush()

            correct_sim.set_wavefunction(basis_state, correct_qb + correct_ctrl_qb)
            test_sim.set_wavefunction(basis_state, test_qb + test_ctrl_qb)
            CNOT | (test_ctrl_qb, test_qb)
            CNOT | (correct_ctrl_qb, correct_qb)

            test_eng.flush()
            correct_eng.flush()

            assert len(correct_dummy_eng.received_commands) == 5
            assert len(test_dummy_eng.received_commands) == 10

            assert correct_eng.backend.cheat()[1] == pytest.approx(test_eng.backend.cheat()[1], rel=1e-12, abs=1e-12)

            All(Measure) | test_qb + test_ctrl_qb
            All(Measure) | correct_qb + correct_ctrl_qb
            test_eng.flush(deallocate_qubits=True)
            correct_eng.flush(deallocate_qubits=True)



# -*- coding: utf-8 -*-
#   Copyright 2020 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Definition of some mathematical quantum operations"""

from projectq.ops import All, X, CNOT
from projectq.meta import Control
from ._gates import AddQuantum, SubtractQuantum


def add_quantum(eng, quint_a, quint_b, carry=None):
    """
    Adds two quantum integers, i.e.,

    |a0...a(n-1)>|b(0)...b(n-1)>|c> -> |a0...a(n-1)>|b+a(0)...b+a(n)>

    (only works if quint_a and quint_b are the same size and carry is a single
    qubit)

    Args:
        eng (MainEngine): ProjectQ MainEngine
        quint_a (list): Quantum register (or list of qubits)
        quint_b (list): Quantum register (or list of qubits)
        carry (list): Carry qubit

    Notes:
        Ancilla: 0, size: 7n-6, toffoli: 2n-1, depth: 5n-3.

    .. rubric:: References

    Quantum addition using ripple carry from: https://arxiv.org/pdf/0910.2530.pdf
    """
    # pylint: disable = pointless-statement

    if len(quint_a) != len(quint_b):
        raise ValueError('quint_a and quint_b must have the same size!')
    if carry and len(carry) != 1:
        raise ValueError('Either no carry bit or a single carry qubit is allowed!')

    n_qubits = len(quint_a) + 1

    for i in range(1, n_qubits - 1):
        CNOT | (quint_a[i], quint_b[i])

    if carry:
        CNOT | (quint_a[n_qubits - 2], carry)

    for j in range(n_qubits - 3, 0, -1):
        CNOT | (quint_a[j], quint_a[j + 1])

    for k in range(0, n_qubits - 2):
        with Control(eng, [quint_a[k], quint_b[k]]):
            X | (quint_a[k + 1])

    if carry:
        with Control(eng, [quint_a[n_qubits - 2], quint_b[n_qubits - 2]]):
            X | carry

    for i in range(n_qubits - 2, 0, -1):  # noqa: E741
        CNOT | (quint_a[i], quint_b[i])
        with Control(eng, [quint_a[i - 1], quint_b[i - 1]]):
            X | quint_a[i]

    for j in range(1, n_qubits - 2):
        CNOT | (quint_a[j], quint_a[j + 1])

    for n_qubits in range(0, n_qubits - 1):
        CNOT | (quint_a[n_qubits], quint_b[n_qubits])


def subtract_quantum(eng, quint_a, quint_b):
    """
    Subtracts two quantum integers, i.e.,

    |a>|b> -> |a>|b-a>

    (only works if quint_a and quint_b are the same size)

    Args:
        eng (MainEngine): ProjectQ MainEngine
        quint_a (list): Quantum register (or list of qubits)
        quint_b (list): Quantum register (or list of qubits)

    Notes:
        Quantum subtraction using bitwise complementation of quantum adder: b-a = (a + b')'. Same as the quantum
        addition circuit except that the steps involving the carry qubit are left out and complement b at the start
        and at the end of the circuit is added.

        Ancilla: 0, size: 9n-8, toffoli: 2n-2, depth: 5n-5.


    .. rubric:: References

    Quantum addition using ripple carry from:
    https://arxiv.org/pdf/0910.2530.pdf
    """
    # pylint: disable = pointless-statement, expression-not-assigned

    if len(quint_a) != len(quint_b):
        raise ValueError('quint_a and quint_b must have the same size!')
    n_qubits = len(quint_a) + 1

    All(X) | quint_b

    for i in range(1, n_qubits - 1):
        CNOT | (quint_a[i], quint_b[i])

    for j in range(n_qubits - 3, 0, -1):
        CNOT | (quint_a[j], quint_a[j + 1])

    for k in range(0, n_qubits - 2):
        with Control(eng, [quint_a[k], quint_b[k]]):
            X | (quint_a[k + 1])

    for i in range(n_qubits - 2, 0, -1):  # noqa: E741
        CNOT | (quint_a[i], quint_b[i])
        with Control(eng, [quint_a[i - 1], quint_b[i - 1]]):
            X | quint_a[i]

    for j in range(1, n_qubits - 2):
        CNOT | (quint_a[j], quint_a[j + 1])

    for n_qubits in range(0, n_qubits - 1):
        CNOT | (quint_a[n_qubits], quint_b[n_qubits])

    All(X) | quint_b


def inverse_add_quantum_carry(eng, quint_a, quint_b):
    """
    Inverse of quantum addition with carry

    Args:
        eng (MainEngine): ProjectQ MainEngine
        quint_a (list): Quantum register (or list of qubits)
        quint_b (list): Quantum register (or list of qubits)
    """
    # pylint: disable = pointless-statement, expression-not-assigned
    # pylint: disable = unused-argument

    if len(quint_a) != len(quint_b[0]):
        raise ValueError('quint_a and quint_b must have the same size!')

    All(X) | quint_b[0]
    X | quint_b[1][0]

    AddQuantum | (quint_a, quint_b[0], quint_b[1])

    All(X) | quint_b[0]
    X | quint_b[1][0]


def comparator(eng, quint_a, quint_b, comp):
    """
    Compares the size of two quantum integers, i.e,

    if a>b: |a>|b>|c> -> |a>|b>|c+1>

    else:   |a>|b>|c> -> |a>|b>|c>

    (only works if quint_a and quint_b are the same size and the comparator is 1 qubit)

    Args:
        eng (MainEngine): ProjectQ MainEngine
        quint_a (list): Quantum register (or list of qubits)
        quint_b (list): Quantum register (or list of qubits)
        comp (Qubit): Comparator qubit

    Notes:
        Comparator flipping a compare qubit by computing the high bit of b-a, which is 1 if and only if a > b. The
        high bit is computed using the first half of circuit in AddQuantum (such that the high bit is written to the
        carry qubit) and then undoing the first half of the circuit. By complementing b at the start and b+a at the
        end the high bit of b-a is calculated.

        Ancilla: 0, size: 8n-3, toffoli: 2n+1, depth: 4n+3.
    """
    # pylint: disable = pointless-statement, expression-not-assigned

    if len(quint_a) != len(quint_b):
        raise ValueError('quint_a and quint_b must have the same size!')
    if len(comp) != 1:
        raise ValueError('Comparator output qubit must be a single qubit!')

    n_qubits = len(quint_a) + 1

    All(X) | quint_b

    for i in range(1, n_qubits - 1):
        CNOT | (quint_a[i], quint_b[i])

    CNOT | (quint_a[n_qubits - 2], comp)

    for j in range(n_qubits - 3, 0, -1):
        CNOT | (quint_a[j], quint_a[j + 1])

    for k in range(0, n_qubits - 2):
        with Control(eng, [quint_a[k], quint_b[k]]):
            X | (quint_a[k + 1])

    with Control(eng, [quint_a[n_qubits - 2], quint_b[n_qubits - 2]]):
        X | comp

    for k in range(0, n_qubits - 2):
        with Control(eng, [quint_a[k], quint_b[k]]):
            X | (quint_a[k + 1])

    for j in range(n_qubits - 3, 0, -1):
        CNOT | (quint_a[j], quint_a[j + 1])

    for i in range(1, n_qubits - 1):
        CNOT | (quint_a[i], quint_b[i])

    All(X) | quint_b


def quantum_conditional_add(eng, quint_a, quint_b, conditional):
    """
    Adds up two quantum integers if conditional is high, i.e.,

    |a>|b>|c> -> |a>|b+a>|c>
    (without a carry out qubit)

    if conditional is low, no operation is performed, i.e.,
    |a>|b>|c> -> |a>|b>|c>

    Args:
        eng (MainEngine): ProjectQ MainEngine
        quint_a (list): Quantum register (or list of qubits)
        quint_b (list): Quantum register (or list of qubits)
        conditional (list): Conditional qubit

    Notes:
        Ancilla: 0, Size: 7n-7, Toffoli: 3n-3, Depth: 5n-3.

    .. rubric:: References

    Quantum Conditional Add from https://arxiv.org/pdf/1609.01241.pdf
    """
    # pylint: disable = pointless-statement, expression-not-assigned

    if len(quint_a) != len(quint_b):
        raise ValueError('quint_a and quint_b must have the same size!')
    if len(conditional) != 1:
        raise ValueError('Conditional qubit must be a single qubit!')

    n_qubits = len(quint_a) + 1

    for i in range(1, n_qubits - 1):
        CNOT | (quint_a[i], quint_b[i])

    for i in range(n_qubits - 2, 1, -1):
        CNOT | (quint_a[i - 1], quint_a[i])

    for k in range(0, n_qubits - 2):
        with Control(eng, [quint_a[k], quint_b[k]]):
            X | (quint_a[k + 1])

    with Control(eng, [quint_a[n_qubits - 2], conditional[0]]):
        X | quint_b[n_qubits - 2]

    for i in range(n_qubits - 2, 0, -1):  # noqa: E741
        with Control(eng, [quint_a[i - 1], quint_b[i - 1]]):
            X | quint_a[i]
        with Control(eng, [quint_a[i - 1], conditional[0]]):
            X | (quint_b[i - 1])

    for j in range(1, n_qubits - 2):
        CNOT | (quint_a[j], quint_a[j + 1])

    for k in range(1, n_qubits - 1):
        CNOT | (quint_a[k], quint_b[k])


def quantum_division(eng, dividend, remainder, divisor):
    """
    Performs restoring integer division, i.e.,

    |dividend>|remainder>|divisor> -> |remainder>|quotient>|divisor>

    (only works if all three qubits are of equal length)

    Args:
        eng (MainEngine): ProjectQ MainEngine
        dividend (list): Quantum register (or list of qubits)
        remainder (list): Quantum register (or list of qubits)
        divisor (list): Quantum register (or list of qubits)

    Notes:
        Ancilla: n, size 16n^2 - 13, toffoli: 5n^2 -5 , depth: 10n^2-6.

    .. rubric:: References

    Quantum Restoring Integer Division from:
    https://arxiv.org/pdf/1609.01241.pdf.
    """
    # The circuit consits of three parts
    # i) leftshift
    # ii) subtraction
    # iii) conditional add operation.

    if not len(dividend) == len(remainder) == len(divisor):
        raise ValueError('Size mismatch in dividend, divisor and remainder!')

    j = len(remainder)
    n_dividend = len(dividend)

    while j != 0:
        combined_reg = []

        combined_reg.append(dividend[n_dividend - 1])

        for i in range(0, n_dividend - 1):
            combined_reg.append(remainder[i])

        SubtractQuantum | (divisor[0:n_dividend], combined_reg)
        CNOT | (combined_reg[n_dividend - 1], remainder[n_dividend - 1])
        with Control(eng, remainder[n_dividend - 1]):
            AddQuantum | (divisor[0:n_dividend], combined_reg)
        X | remainder[n_dividend - 1]

        remainder.insert(0, dividend[n_dividend - 1])
        dividend.insert(0, remainder[n_dividend])
        del remainder[n_dividend]
        del dividend[n_dividend]

        j -= 1


def inverse_quantum_division(eng, remainder, quotient, divisor):
    """
    Performs the inverse of a restoring integer division, i.e.,

    |remainder>|quotient>|divisor> ->  |dividend>|remainder(0)>|divisor>

    Args:
        eng (MainEngine): ProjectQ MainEngine
        dividend (list): Quantum register (or list of qubits)
        remainder (list): Quantum register (or list of qubits)
        divisor (list): Quantum register (or list of qubits)
    """
    if not len(quotient) == len(remainder) == len(divisor):
        raise ValueError('Size mismatch in quotient, divisor and remainder!')

    j = 0
    n_quotient = len(quotient)

    while j != n_quotient:
        X | quotient[0]
        with Control(eng, quotient[0]):
            SubtractQuantum | (divisor, remainder)
        CNOT | (remainder[-1], quotient[0])

        AddQuantum | (divisor, remainder)

        remainder.insert(n_quotient, quotient[0])
        quotient.insert(n_quotient, remainder[0])
        del remainder[0]
        del quotient[0]
        j += 1


def quantum_conditional_add_carry(eng, quint_a, quint_b, ctrl, z):  # pylint: disable=invalid-name
    """
    Adds up two quantum integers if the control qubit is |1>, i.e.,

    |a>|b>|ctrl>|z(0)z(1)> -> |a>|s(0)...s(n-1)>|ctrl>|s(n)z(1)>
    (where s denotes the sum of a and b)

    If the control qubit is |0> no operation is performed:

    |a>|b>|ctrl>|z(0)z(1)> -> |a>|b>|ctrl>|z(0)z(1)>

    (only works if quint_a and quint_b are of the same size, ctrl is a
    single qubit and z is a quantum register with 2 qubits.

    Args:
        eng (MainEngine): ProjectQ MainEngine
        quint_a (list): Quantum register (or list of qubits)
        quint_b (list): Quantum register (or list of qubits)
        ctrl (list): Control qubit
        z (list): Quantum register with 2 qubits

    Notes:
        Ancilla: 2, size: 7n - 4, toffoli: 3n + 2, depth: 5n.

    .. rubric:: References

    Quantum conditional add with no input carry from: https://arxiv.org/pdf/1706.05113.pdf
    """
    if len(quint_a) != len(quint_b):
        raise ValueError('quint_a and quint_b must have the same size!')
    if len(ctrl) != 1:
        raise ValueError('Only a single control qubit is allowed!')
    if len(z) != 2:
        raise ValueError('Z quantum register must have 2 qubits!')

    n_a = len(quint_a)

    for i in range(1, n_a):
        CNOT | (quint_a[i], quint_b[i])

    with Control(eng, [quint_a[n_a - 1], ctrl[0]]):
        X | z[0]

    for j in range(n_a - 2, 0, -1):
        CNOT | (quint_a[j], quint_a[j + 1])

    for k in range(0, n_a - 1):
        with Control(eng, [quint_b[k], quint_a[k]]):
            X | quint_a[k + 1]

    with Control(eng, [quint_b[n_a - 1], quint_a[n_a - 1]]):
        X | z[1]

    with Control(eng, [ctrl[0], z[1]]):
        X | z[0]

    with Control(eng, [quint_b[n_a - 1], quint_a[n_a - 1]]):
        X | z[1]

    for i in range(n_a - 1, 0, -1):  # noqa: E741
        with Control(eng, [ctrl[0], quint_a[i]]):
            X | quint_b[i]
        with Control(eng, [quint_a[i - 1], quint_b[i - 1]]):
            X | quint_a[i]

    with Control(eng, [quint_a[0], ctrl[0]]):
        X | quint_b[0]

    for j in range(1, n_a - 1):
        CNOT | (quint_a[j], quint_a[j + 1])

    for n_a in range(1, n_a):
        CNOT | (quint_a[n_a], quint_b[n_a])


def quantum_multiplication(eng, quint_a, quint_b, product):
    """
    Multiplies two quantum integers, i.e,

    |a>|b>|0> -> |a>|b>|a*b>

    (only works if quint_a and quint_b are of the same size, n qubits and product has size 2n+1).

    Args:
        eng (MainEngine): ProjectQ MainEngine
        quint_a (list): Quantum register (or list of qubits)
        quint_b (list): Quantum register (or list of qubits)
        product (list): Quantum register (or list of qubits) storing
            the result

    Notes:
        Ancilla: 2n + 1, size: 7n^2 - 9n + 4, toffoli: 5n^2 - 4n, depth: 3n^2 - 2.

    .. rubric:: References

    Quantum multiplication from: https://arxiv.org/abs/1706.05113.

    """
    n_a = len(quint_a)

    if len(quint_a) != len(quint_b):
        raise ValueError('quint_a and quint_b must have the same size!')
    if len(product) != ((2 * n_a) + 1):
        raise ValueError('product size must be 2*n + 1')

    for i in range(0, n_a):
        with Control(eng, [quint_a[i], quint_b[0]]):
            X | product[i]

    with Control(eng, quint_b[1]):
        AddQuantum | (
            quint_a[0 : (n_a - 1)],  # noqa: E203
            product[1:n_a],
            [product[n_a + 1], product[n_a + 2]],
        )

    for j in range(2, n_a):
        with Control(eng, quint_b[j]):
            AddQuantum | (
                quint_a[0 : (n_a - 1)],  # noqa: E203
                product[(0 + j) : (n_a - 1 + j)],  # noqa: E203
                [product[n_a + j], product[n_a + j + 1]],
            )


def inverse_quantum_multiplication(eng, quint_a, quint_b, product):
    """
    Inverse of the multiplication of two quantum integers, i.e,

    |a>|b>|a*b> -> |a>|b>|0>

    (only works if quint_a and quint_b are of the same size, n qubits and product has size 2n+1)

    Args:
        eng (MainEngine): ProjectQ MainEngine
        quint_a (list): Quantum register (or list of qubits)
        quint_b (list): Quantum register (or list of qubits)
        product (list): Quantum register (or list of qubits) storing the result

    """
    n_a = len(quint_a)

    if len(quint_a) != len(quint_b):
        raise ValueError('quint_a and quint_b must have the same size!')
    if len(product) != ((2 * n_a) + 1):
        raise ValueError('product size must be 2*n + 1')

    for j in range(2, n_a):
        with Control(eng, quint_b[j]):
            SubtractQuantum | (
                quint_a[0 : (n_a - 1)],  # noqa: E203
                product[(0 + j) : (n_a - 1 + j)],  # noqa: E203
                [product[n_a + j], product[n_a + j + 1]],
            )
    for i in range(0, n_a):
        with Control(eng, [quint_a[i], quint_b[0]]):
            X | product[i]

    with Control(eng, quint_b[1]):
        SubtractQuantum | (
            quint_a[0 : (n_a - 1)],  # noqa: E203
            product[1:n_a],
            [product[n_a + 1], product[n_a + 2]],
        )



# -*- coding: utf-8 -*-
#   Copyright 2021 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
""" Test for projectq.backends._awsbraket._awsbraket_boto3_client.py """

import pytest

from ._awsbraket_boto3_client_test_fixtures import *  # noqa: F401,F403

# ==============================================================================

_has_boto3 = True
try:
    import botocore
    from projectq.backends._awsbraket import _awsbraket_boto3_client
except ImportError:
    _has_boto3 = False

has_boto3 = pytest.mark.skipif(not _has_boto3, reason="boto3 package is not installed")

# ==============================================================================


@has_boto3
def test_show_devices(mocker, show_devices_setup):
    creds, search_value, device_value, devicelist_result = show_devices_setup

    mock_boto3_client = mocker.MagicMock(spec=['search_devices', 'get_device'])
    mock_boto3_client.search_devices.return_value = search_value
    mock_boto3_client.get_device.return_value = device_value
    mocker.patch('boto3.client', return_value=mock_boto3_client, autospec=True)

    devicelist = _awsbraket_boto3_client.show_devices(credentials=creds)
    assert devicelist == devicelist_result


# ==============================================================================

completed_value = {
    'deviceArn': 'arndevice',
    'deviceParameters': 'parameters',
    'failureReason': 'None',
    'outputS3Bucket': 'amazon-braket-bucket',
    'outputS3Directory': 'complete/directory',
    'quantumTaskArn': 'arntask',
    'shots': 123,
    'status': 'COMPLETED',
    'tags': {'tagkey': 'tagvalue'},
}

failed_value = {
    'failureReason': 'This is a failure reason',
    'outputS3Bucket': 'amazon-braket-bucket',
    'outputS3Directory': 'complete/directory',
    'status': 'FAILED',
}

cancelling_value = {
    'failureReason': 'None',
    'outputS3Bucket': 'amazon-braket-bucket',
    'outputS3Directory': 'complete/directory',
    'status': 'CANCELLING',
}

other_value = {
    'failureReason': 'None',
    'outputS3Bucket': 'amazon-braket-bucket',
    'outputS3Directory': 'complete/directory',
    'status': 'OTHER',
}

# ------------------------------------------------------------------------------


@has_boto3
@pytest.mark.parametrize(
    "var_status, var_result",
    [
        ('completed', completed_value),
        ('failed', failed_value),
        ('cancelling', cancelling_value),
        ('other', other_value),
    ],
)
def test_retrieve(mocker, var_status, var_result, retrieve_setup):
    arntask, creds, device_value, res_completed, results_dict = retrieve_setup

    mock_boto3_client = mocker.MagicMock(spec=['get_quantum_task', 'get_device', 'get_object'])
    mock_boto3_client.get_quantum_task.return_value = var_result
    mock_boto3_client.get_device.return_value = device_value
    mock_boto3_client.get_object.return_value = results_dict
    mocker.patch('boto3.client', return_value=mock_boto3_client, autospec=True)

    if var_status == 'completed':
        res = _awsbraket_boto3_client.retrieve(credentials=creds, task_arn=arntask)
        assert res == res_completed
    else:
        with pytest.raises(Exception) as exinfo:
            _awsbraket_boto3_client.retrieve(credentials=creds, task_arn=arntask, num_retries=2)
        print(exinfo.value)
        if var_status == 'failed':
            assert (
                str(exinfo.value)
                == "Error while running the code: FAILED. \
The failure reason was: This is a failure reason."
            )

        if var_status == 'cancelling':
            assert str(exinfo.value) == "The job received a CANCEL operation: CANCELLING."
        if var_status == 'other':
            assert (
                str(exinfo.value)
                == "Timeout. The Arn of your submitted job \
is arn:aws:braket:us-east-1:id:taskuuid \
and the status of the job is OTHER."
            )


# ==============================================================================


@has_boto3
def test_retrieve_devicetypes(mocker, retrieve_devicetypes_setup):
    (
        arntask,
        creds,
        device_value,
        results_dict,
        res_completed,
    ) = retrieve_devicetypes_setup

    mock_boto3_client = mocker.MagicMock(spec=['get_quantum_task', 'get_device', 'get_object'])
    mock_boto3_client.get_quantum_task.return_value = completed_value
    mock_boto3_client.get_device.return_value = device_value
    mock_boto3_client.get_object.return_value = results_dict
    mocker.patch('boto3.client', return_value=mock_boto3_client, autospec=True)

    res = _awsbraket_boto3_client.retrieve(credentials=creds, task_arn=arntask)
    assert res == res_completed


# ==============================================================================


@has_boto3
def test_send_too_many_qubits(mocker, send_too_many_setup):
    (creds, s3_folder, search_value, device_value, info_too_much) = send_too_many_setup

    mock_boto3_client = mocker.MagicMock(spec=['search_devices', 'get_device'])
    mock_boto3_client.search_devices.return_value = search_value
    mock_boto3_client.get_device.return_value = device_value
    mocker.patch('boto3.client', return_value=mock_boto3_client, autospec=True)

    with pytest.raises(_awsbraket_boto3_client.DeviceTooSmall):
        _awsbraket_boto3_client.send(info_too_much, device='name2', credentials=creds, s3_folder=s3_folder)


# ==============================================================================


@has_boto3
@pytest.mark.parametrize(
    "var_status, var_result",
    [
        ('completed', completed_value),
        ('failed', failed_value),
        ('cancelling', cancelling_value),
        ('other', other_value),
    ],
)
def test_send_real_device_online_verbose(mocker, var_status, var_result, real_device_online_setup):

    (
        qtarntask,
        creds,
        s3_folder,
        info,
        search_value,
        device_value,
        res_completed,
        results_dict,
    ) = real_device_online_setup

    mock_boto3_client = mocker.MagicMock(
        spec=['search_devices', 'get_device', 'create_quantum_task', 'get_quantum_task', 'get_object']
    )
    mock_boto3_client.search_devices.return_value = search_value
    mock_boto3_client.get_device.return_value = device_value
    mock_boto3_client.create_quantum_task.return_value = qtarntask
    mock_boto3_client.get_quantum_task.return_value = var_result
    mock_boto3_client.get_object.return_value = results_dict
    mocker.patch('boto3.client', return_value=mock_boto3_client, autospec=True)

    # This is a ficticios situation because the job will be always queued
    # at the beginning. After that the status will change at some point in time
    # If the status change while the _get_result loop with num_retries, is
    # active the result will change. We mock this using some preconfigured
    # statuses in var_status for the tests

    if var_status == 'completed':
        res = _awsbraket_boto3_client.send(info, device='name2', credentials=creds, s3_folder=s3_folder, verbose=True)
        assert res == res_completed
    else:
        with pytest.raises(Exception) as exinfo:
            _awsbraket_boto3_client.send(
                info,
                device='name2',
                credentials=creds,
                s3_folder=s3_folder,
                verbose=True,
                num_retries=2,
            )
        print(exinfo.value)
        if var_status == 'failed':
            assert (
                str(exinfo.value)
                == "Error while running the code: FAILED. The failure \
reason was: This is a failure reason."
            )

        if var_status == 'cancelling':
            assert str(exinfo.value) == "The job received a CANCEL operation: CANCELLING."
        if var_status == 'other':
            assert (
                str(exinfo.value)
                == "Timeout. The Arn of your submitted job \
is arn:aws:braket:us-east-1:id:taskuuid \
and the status of the job is OTHER."
            )


# ==============================================================================


@has_boto3
@pytest.mark.parametrize(
    "var_error",
    [
        ('AccessDeniedException'),
        ('DeviceOfflineException'),
        ('InternalServiceException'),
        ('ServiceQuotaExceededException'),
        ('ValidationException'),
    ],
)
def test_send_that_errors_are_caught(mocker, var_error, send_that_error_setup):
    creds, s3_folder, info, search_value, device_value = send_that_error_setup

    mock_boto3_client = mocker.MagicMock(spec=['search_devices', 'get_device', 'create_quantum_task'])
    mock_boto3_client.search_devices.return_value = search_value
    mock_boto3_client.get_device.return_value = device_value
    mock_boto3_client.create_quantum_task.side_effect = botocore.exceptions.ClientError(
        {"Error": {"Code": var_error, "Message": "Msg error for " + var_error}},
        "create_quantum_task",
    )
    mocker.patch('boto3.client', return_value=mock_boto3_client, autospec=True)

    with pytest.raises(botocore.exceptions.ClientError):
        _awsbraket_boto3_client.send(info, device='name2', credentials=creds, s3_folder=s3_folder, num_retries=2)

    with pytest.raises(_awsbraket_boto3_client.DeviceOfflineError):
        _awsbraket_boto3_client.send(
            info,
            device='unknown',
            credentials=creds,
            s3_folder=s3_folder,
            num_retries=2,
        )


# ==============================================================================


@has_boto3
@pytest.mark.parametrize("var_error", [('ResourceNotFoundException')])
def test_retrieve_error_arn_not_exist(mocker, var_error, arntask, creds):

    mock_boto3_client = mocker.MagicMock(spec=['get_quantum_task'])
    mock_boto3_client.get_quantum_task.side_effect = botocore.exceptions.ClientError(
        {"Error": {"Code": var_error, "Message": "Msg error for " + var_error}},
        "get_quantum_task",
    )
    mocker.patch('boto3.client', return_value=mock_boto3_client, autospec=True)

    with pytest.raises(botocore.exceptions.ClientError):
        _awsbraket_boto3_client.retrieve(credentials=creds, task_arn=arntask)


# ==============================================================================



# -*- coding: utf-8 -*-
# pylint: skip-file

import projectq.setups.ibm
from projectq.ops import H, Measure
from projectq import MainEngine
from projectq.backends import IBMBackend

# create a main compiler engine
eng = MainEngine(IBMBackend(), engine_list=projectq.setups.ibm.get_engine_list())

# allocate one qubit
q1 = eng.allocate_qubit()

# put it in superposition
H | q1

# measure
Measure | q1

eng.flush()
# print the result:
print("Measured: {}".format(int(q1)))



# -*- coding: utf-8 -*-
#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"Tests for projectq.setups.decompositions.phaseestimation.py."

import cmath
import numpy as np
from flaky import flaky
import pytest

from projectq.backends import Simulator
from projectq.cengines import (
    AutoReplacer,
    DecompositionRuleSet,
    MainEngine,
)

from projectq.ops import X, H, All, Measure, Tensor, Ph, CNOT, StatePreparation, QPE

from projectq.setups.decompositions import phaseestimation as pe
from projectq.setups.decompositions import qft2crandhadamard as dqft
import projectq.setups.decompositions.stateprep2cnot as stateprep2cnot
import projectq.setups.decompositions.uniformlycontrolledr2cnot as ucr2cnot


@flaky(max_runs=5, min_passes=2)
def test_simple_test_X_eigenvectors():
    rule_set = DecompositionRuleSet(modules=[pe, dqft])
    eng = MainEngine(
        backend=Simulator(),
        engine_list=[
            AutoReplacer(rule_set),
        ],
    )
    results = np.array([])
    for i in range(150):
        autovector = eng.allocate_qureg(1)
        X | autovector
        H | autovector
        unit = X
        ancillas = eng.allocate_qureg(1)
        QPE(unit) | (ancillas, autovector)
        All(Measure) | ancillas
        fasebinlist = [int(q) for q in ancillas]
        fasebin = ''.join(str(j) for j in fasebinlist)
        faseint = int(fasebin, 2)
        phase = faseint / (2.0 ** (len(ancillas)))
        results = np.append(results, phase)
        All(Measure) | autovector
        eng.flush()

    num_phase = (results == 0.5).sum()
    assert num_phase / 100.0 >= 0.35, "Statistics phase calculation are not correct (%f vs. %f)" % (
        num_phase / 100.0,
        0.35,
    )


@flaky(max_runs=5, min_passes=2)
def test_Ph_eigenvectors():
    rule_set = DecompositionRuleSet(modules=[pe, dqft])
    eng = MainEngine(
        backend=Simulator(),
        engine_list=[
            AutoReplacer(rule_set),
        ],
    )
    results = np.array([])
    for i in range(150):
        autovector = eng.allocate_qureg(1)
        theta = cmath.pi * 2.0 * 0.125
        unit = Ph(theta)
        ancillas = eng.allocate_qureg(3)
        QPE(unit) | (ancillas, autovector)
        All(Measure) | ancillas
        fasebinlist = [int(q) for q in ancillas]
        fasebin = ''.join(str(j) for j in fasebinlist)
        faseint = int(fasebin, 2)
        phase = faseint / (2.0 ** (len(ancillas)))
        results = np.append(results, phase)
        All(Measure) | autovector
        eng.flush()

    num_phase = (results == 0.125).sum()
    assert num_phase / 100.0 >= 0.35, "Statistics phase calculation are not correct (%f vs. %f)" % (
        num_phase / 100.0,
        0.35,
    )


def two_qubit_gate(system_q, time):
    CNOT | (system_q[0], system_q[1])
    Ph(2.0 * cmath.pi * (time * 0.125)) | system_q[1]
    CNOT | (system_q[0], system_q[1])


@flaky(max_runs=5, min_passes=2)
def test_2qubitsPh_andfunction_eigenvectors():
    rule_set = DecompositionRuleSet(modules=[pe, dqft])
    eng = MainEngine(
        backend=Simulator(),
        engine_list=[
            AutoReplacer(rule_set),
        ],
    )
    results = np.array([])
    for i in range(150):
        autovector = eng.allocate_qureg(2)
        X | autovector[0]
        ancillas = eng.allocate_qureg(3)
        QPE(two_qubit_gate) | (ancillas, autovector)
        All(Measure) | ancillas
        fasebinlist = [int(q) for q in ancillas]
        fasebin = ''.join(str(j) for j in fasebinlist)
        faseint = int(fasebin, 2)
        phase = faseint / (2.0 ** (len(ancillas)))
        results = np.append(results, phase)
        All(Measure) | autovector
        eng.flush()

    num_phase = (results == 0.125).sum()
    assert num_phase / 100.0 >= 0.34, "Statistics phase calculation are not correct (%f vs. %f)" % (
        num_phase / 100.0,
        0.34,
    )


def test_X_no_eigenvectors():
    rule_set = DecompositionRuleSet(modules=[pe, dqft, stateprep2cnot, ucr2cnot])
    eng = MainEngine(
        backend=Simulator(),
        engine_list=[
            AutoReplacer(rule_set),
        ],
    )
    results = np.array([])
    results_plus = np.array([])
    results_minus = np.array([])
    for i in range(100):
        autovector = eng.allocate_qureg(1)
        amplitude0 = (np.sqrt(2) + np.sqrt(6)) / 4.0
        amplitude1 = (np.sqrt(2) - np.sqrt(6)) / 4.0
        StatePreparation([amplitude0, amplitude1]) | autovector
        unit = X
        ancillas = eng.allocate_qureg(1)
        QPE(unit) | (ancillas, autovector)
        All(Measure) | ancillas
        fasebinlist = [int(q) for q in ancillas]
        fasebin = ''.join(str(j) for j in fasebinlist)
        faseint = int(fasebin, 2)
        phase = faseint / (2.0 ** (len(ancillas)))
        results = np.append(results, phase)
        Tensor(H) | autovector
        if np.allclose(phase, 0.0, rtol=1e-1):
            results_plus = np.append(results_plus, phase)
            All(Measure) | autovector
            autovector_result = int(autovector)
            assert autovector_result == 0
        elif np.allclose(phase, 0.5, rtol=1e-1):
            results_minus = np.append(results_minus, phase)
            All(Measure) | autovector
            autovector_result = int(autovector)
            assert autovector_result == 1
        eng.flush()

    total = len(results_plus) + len(results_minus)
    plus_probability = len(results_plus) / 100.0
    assert total == pytest.approx(100, abs=5)
    assert plus_probability == pytest.approx(
        1.0 / 4.0, abs=1e-1
    ), "Statistics on |+> probability are not correct (%f vs. %f)" % (
        plus_probability,
        1.0 / 4.0,
    )


def test_string():
    unit = X
    gate = QPE(unit)
    assert str(gate) == "QPE(X)"



# -*- coding: utf-8 -*-
#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"Tests for projectq.setups.decompositions.cnu2toffoliandcu."

import pytest

from projectq.backends import Simulator
from projectq.cengines import (
    AutoReplacer,
    DecompositionRuleSet,
    DummyEngine,
    InstructionFilter,
    MainEngine,
)
from projectq.meta import Control
from projectq.ops import (
    All,
    ClassicalInstructionGate,
    Measure,
    Ph,
    QFT,
    Rx,
    Ry,
    X,
    XGate,
)

from . import cnu2toffoliandcu


def test_recognize_correct_gates():
    saving_backend = DummyEngine(save_commands=True)
    eng = MainEngine(backend=saving_backend)
    qubit = eng.allocate_qubit()
    ctrl_qureg = eng.allocate_qureg(2)
    ctrl_qureg2 = eng.allocate_qureg(3)
    eng.flush()
    with Control(eng, ctrl_qureg):
        Ph(0.1) | qubit
        Ry(0.2) | qubit
    with Control(eng, ctrl_qureg2):
        QFT | qubit + ctrl_qureg
        X | qubit
    eng.flush()  # To make sure gates arrive before deallocate gates
    eng.flush(deallocate_qubits=True)
    # Don't test initial 6 allocate and flush and trailing deallocate
    # and two flush gates.
    for cmd in saving_backend.received_commands[7:-8]:
        assert cnu2toffoliandcu._recognize_CnU(cmd)


def test_recognize_incorrect_gates():
    saving_backend = DummyEngine(save_commands=True)
    eng = MainEngine(backend=saving_backend)
    qubit = eng.allocate_qubit()
    ctrl_qubit = eng.allocate_qubit()
    ctrl_qureg = eng.allocate_qureg(2)
    eng.flush()
    with Control(eng, ctrl_qubit):
        Rx(0.3) | qubit
    X | qubit
    with Control(eng, ctrl_qureg):
        X | qubit
    eng.flush(deallocate_qubits=True)
    for cmd in saving_backend.received_commands:
        assert not cnu2toffoliandcu._recognize_CnU(cmd)


def _decomp_gates(eng, cmd):
    g = cmd.gate
    if isinstance(g, ClassicalInstructionGate):
        return True
    if len(cmd.control_qubits) <= 1:
        return True
    if len(cmd.control_qubits) == 2 and isinstance(cmd.gate, XGate):
        return True
    return False


def test_decomposition():
    for basis_state_index in range(0, 16):
        basis_state = [0] * 16
        basis_state[basis_state_index] = 1.0
        correct_dummy_eng = DummyEngine(save_commands=True)
        correct_eng = MainEngine(backend=Simulator(), engine_list=[correct_dummy_eng])
        rule_set = DecompositionRuleSet(modules=[cnu2toffoliandcu])
        test_dummy_eng = DummyEngine(save_commands=True)
        test_eng = MainEngine(
            backend=Simulator(),
            engine_list=[
                AutoReplacer(rule_set),
                InstructionFilter(_decomp_gates),
                test_dummy_eng,
            ],
        )
        test_sim = test_eng.backend
        correct_sim = correct_eng.backend
        correct_qb = correct_eng.allocate_qubit()
        correct_ctrl_qureg = correct_eng.allocate_qureg(3)
        correct_eng.flush()
        test_qb = test_eng.allocate_qubit()
        test_ctrl_qureg = test_eng.allocate_qureg(3)
        test_eng.flush()

        correct_sim.set_wavefunction(basis_state, correct_qb + correct_ctrl_qureg)
        test_sim.set_wavefunction(basis_state, test_qb + test_ctrl_qureg)

        with Control(test_eng, test_ctrl_qureg[:2]):
            Rx(0.4) | test_qb
        with Control(test_eng, test_ctrl_qureg):
            Ry(0.6) | test_qb
        with Control(test_eng, test_ctrl_qureg):
            X | test_qb

        with Control(correct_eng, correct_ctrl_qureg[:2]):
            Rx(0.4) | correct_qb
        with Control(correct_eng, correct_ctrl_qureg):
            Ry(0.6) | correct_qb
        with Control(correct_eng, correct_ctrl_qureg):
            X | correct_qb

        test_eng.flush()
        correct_eng.flush()

        assert len(correct_dummy_eng.received_commands) == 9
        assert len(test_dummy_eng.received_commands) == 25

        for fstate in range(16):
            binary_state = format(fstate, '04b')
            test = test_sim.get_amplitude(binary_state, test_qb + test_ctrl_qureg)
            correct = correct_sim.get_amplitude(binary_state, correct_qb + correct_ctrl_qureg)
            assert correct == pytest.approx(test, rel=1e-12, abs=1e-12)

        All(Measure) | test_qb + test_ctrl_qureg
        All(Measure) | correct_qb + correct_ctrl_qureg
        test_eng.flush(deallocate_qubits=True)
        correct_eng.flush(deallocate_qubits=True)



# -*- coding: utf-8 -*-
#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"Tests for projectq.setups.decompositions.h2rx.py"

import pytest

from projectq import MainEngine
from projectq.backends import Simulator
from projectq.cengines import (
    AutoReplacer,
    DecompositionRuleSet,
    DummyEngine,
    InstructionFilter,
)
from projectq.meta import Control
from projectq.ops import Measure, H, HGate

from . import h2rx


def test_recognize_correct_gates():
    """Test that recognize_HNoCtrl recognizes ctrl qubits"""
    saving_backend = DummyEngine(save_commands=True)
    eng = MainEngine(backend=saving_backend)
    qubit = eng.allocate_qubit()
    ctrl_qubit = eng.allocate_qubit()
    eng.flush()
    H | qubit
    with Control(eng, ctrl_qubit):
        H | qubit
    eng.flush(deallocate_qubits=True)
    assert h2rx._recognize_HNoCtrl(saving_backend.received_commands[3])
    assert not h2rx._recognize_HNoCtrl(saving_backend.received_commands[4])


def h_decomp_gates(eng, cmd):
    """Test that cmd.gate is a gate of class HGate"""
    g = cmd.gate
    if isinstance(g, HGate):  # H is just a shortcut to HGate
        return False
    else:
        return True


# ------------test_decomposition function-------------#
# Creates two engines, correct_eng and test_eng.
# correct_eng implements H gate.
# test_eng implements the decomposition of the H gate.
# correct_qb and test_qb represent results of these two engines, respectively.
#
# The decomposition in this case only produces the same state as H up to a
# global phase.
# test_vector and correct_vector represent the final wave states of correct_qb
# and test_qb.
# The dot product of correct_vector and test_vector should have absolute value
# 1, if the two vectors are the same up to a global phase.


def test_decomposition():
    """Test that this decomposition of H produces correct amplitudes

    Function tests each DecompositionRule in
    h2rx.all_defined_decomposition_rules
    """
    decomposition_rule_list = h2rx.all_defined_decomposition_rules
    for rule in decomposition_rule_list:
        for basis_state_index in range(2):
            basis_state = [0] * 2
            basis_state[basis_state_index] = 1.0

            correct_dummy_eng = DummyEngine(save_commands=True)
            correct_eng = MainEngine(backend=Simulator(), engine_list=[correct_dummy_eng])

            rule_set = DecompositionRuleSet(rules=[rule])
            test_dummy_eng = DummyEngine(save_commands=True)
            test_eng = MainEngine(
                backend=Simulator(),
                engine_list=[
                    AutoReplacer(rule_set),
                    InstructionFilter(h_decomp_gates),
                    test_dummy_eng,
                ],
            )

            correct_qb = correct_eng.allocate_qubit()
            correct_eng.flush()
            test_qb = test_eng.allocate_qubit()
            test_eng.flush()

            correct_eng.backend.set_wavefunction(basis_state, correct_qb)
            test_eng.backend.set_wavefunction(basis_state, test_qb)

            H | correct_qb
            H | test_qb

            correct_eng.flush()
            test_eng.flush()

            assert H in (cmd.gate for cmd in correct_dummy_eng.received_commands)
            assert H not in (cmd.gate for cmd in test_dummy_eng.received_commands)

            assert correct_eng.backend.cheat()[1] == pytest.approx(test_eng.backend.cheat()[1], rel=1e-12, abs=1e-12)

            Measure | test_qb
            Measure | correct_qb



# -*- coding: utf-8 -*-
#   Copyright 2018 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""
Defines a setup to compile to qubits placed in a linear chain or a circle.

It provides the `engine_list` for the `MainEngine`. This engine list contains an AutoReplacer with most of the gate
decompositions of ProjectQ, which are used to decompose a circuit into only two qubit gates and arbitrary single qubit
gates. ProjectQ's LinearMapper is then used to introduce the necessary Swap operations to route interacting qubits
next to each other. This setup allows to choose the final gate set (with some limitations).
"""

from projectq.cengines import LinearMapper
from projectq.ops import CNOT, Swap

from ._utils import get_engine_list_linear_grid_base


def get_engine_list(num_qubits, cyclic=False, one_qubit_gates="any", two_qubit_gates=(CNOT, Swap)):
    """
    Returns an engine list to compile to a linear chain of qubits.

    Note:
        If you choose a new gate set for which the compiler does not yet have standard rules, it raises an
        `NoGateDecompositionError` or a `RuntimeError: maximum recursion depth exceeded...`. Also note that even the
        gate sets which work might not yet be optimized. So make sure to double check and potentially extend the
        decomposition rules.  This implemention currently requires that the one qubit gates must contain Rz and at
        least one of {Ry(best), Rx, H} and the two qubit gate must contain CNOT (recommended) or CZ.

    Note:
        Classical instructions gates such as e.g. Flush and Measure are automatically allowed.

    Example:
        get_engine_list(num_qubits=10, cyclic=False,
                        one_qubit_gates=(Rz, Ry, Rx, H),
                        two_qubit_gates=(CNOT,))

    Args:
        num_qubits(int): Number of qubits in the chain
        cyclic(bool): If a circle or not. Default is False
        one_qubit_gates: "any" allows any one qubit gate, otherwise provide a tuple of the allowed gates. If the gates
                         are instances of a class (e.g. X), it allows all gates which are equal to it. If the gate is
                         a class (Rz), it allows all instances of this class. Default is "any"
        two_qubit_gates: "any" allows any two qubit gate, otherwise provide a tuple of the allowed gates. If the gates
                         are instances of a class (e.g. CNOT), it allows all gates which are equal to it. If the gate
                         is a class, it allows all instances of this class.  Default is (CNOT, Swap).
    Raises:
        TypeError: If input is for the gates is not "any" or a tuple.

    Returns:
        A list of suitable compiler engines.
    """
    return get_engine_list_linear_grid_base(
        LinearMapper(num_qubits=num_qubits, cyclic=cyclic), one_qubit_gates, two_qubit_gates
    )



# -*- coding: utf-8 -*-
#   Copyright 2021 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
Tests for the controlstate decomposition rule.
"""

from projectq import MainEngine
from projectq.cengines import DummyEngine, AutoReplacer, InstructionFilter, DecompositionRuleSet
from projectq.meta import Control, has_negative_control
from projectq.ops import X
from projectq.setups.decompositions import controlstate, cnot2cz


def filter_func(eng, cmd):
    if has_negative_control(cmd):
        return False
    return True


def test_controlstate_priority():
    saving_backend = DummyEngine(save_commands=True)
    rule_set = DecompositionRuleSet(modules=[cnot2cz, controlstate])
    eng = MainEngine(backend=saving_backend, engine_list=[AutoReplacer(rule_set), InstructionFilter(filter_func)])
    qubit1 = eng.allocate_qubit()
    qubit2 = eng.allocate_qubit()
    qubit3 = eng.allocate_qubit()
    with Control(eng, qubit2, ctrl_state='0'):
        X | qubit1
    with Control(eng, qubit3, ctrl_state='1'):
        X | qubit1
    eng.flush()

    assert len(saving_backend.received_commands) == 8
    for cmd in saving_backend.received_commands:
        assert not has_negative_control(cmd)



# -*- coding: utf-8 -*-
# pylint: skip-file

from projectq.ops import H, Measure
from projectq import MainEngine

# create a main compiler engine
eng = MainEngine()

# allocate one qubit
q1 = eng.allocate_qubit()

# put it in superposition
H | q1

# measure
Measure | q1

eng.flush()
# print the result:
print("Measured: {}".format(int(q1)))



# -*- coding: utf-8 -*-
#   Copyright 2018 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""
Some utility functions common to some setups
"""
import inspect

from projectq.cengines import (
    AutoReplacer,
    DecompositionRuleSet,
    InstructionFilter,
    LocalOptimizer,
    TagRemover,
)
from projectq.ops import ClassicalInstructionGate, CNOT, ControlledGate, Swap, QFT, get_inverse, BasicMathGate
import projectq.libs.math
import projectq.setups.decompositions


def one_and_two_qubit_gates(eng, cmd):  # pylint: disable=unused-argument
    """
    Filter out 1- and 2-qubit gates.
    """
    all_qubits = [qb for qureg in cmd.all_qubits for qb in qureg]
    if isinstance(cmd.gate, ClassicalInstructionGate):
        # This is required to allow Measure, Allocate, Deallocate, Flush
        return True
    if eng.next_engine.is_available(cmd):
        return True
    if len(all_qubits) <= 2:
        return True
    return False


def high_level_gates(eng, cmd):  # pylint: disable=unused-argument
    """
    Remove any MathGates.
    """
    gate = cmd.gate
    if eng.next_engine.is_available(cmd):
        return True
    if gate == QFT or get_inverse(gate) == QFT or gate == Swap:
        return True
    if isinstance(gate, BasicMathGate):
        return False
    return True


def get_engine_list_linear_grid_base(mapper, one_qubit_gates="any", two_qubit_gates=(CNOT, Swap)):
    """
    Returns an engine list to compile to a 2-D grid of qubits.

    Note:
        If you choose a new gate set for which the compiler does not yet have standard rules, it raises an
        `NoGateDecompositionError` or a `RuntimeError: maximum recursion depth exceeded...`. Also note that even the
        gate sets which work might not yet be optimized. So make sure to double check and potentially extend the
        decomposition rules.  This implemention currently requires that the one qubit gates must contain Rz and at
        least one of {Ry(best), Rx, H} and the two qubit gate must contain CNOT (recommended) or CZ.

    Note:
        Classical instructions gates such as e.g. Flush and Measure are automatically allowed.

    Example:
        get_engine_list(num_rows=2, num_columns=3,
                        one_qubit_gates=(Rz, Ry, Rx, H),
                        two_qubit_gates=(CNOT,))

    Args:
        num_rows(int): Number of rows in the grid
        num_columns(int): Number of columns in the grid.
        one_qubit_gates: "any" allows any one qubit gate, otherwise provide a tuple of the allowed gates. If the gates
                         are instances of a class (e.g. X), it allows all gates which are equal to it. If the gate is
                         a class (Rz), it allows all instances of this class. Default is "any"
        two_qubit_gates: "any" allows any two qubit gate, otherwise provide a tuple of the allowed gates. If the gates
                         are instances of a class (e.g. CNOT), it allows all gates which are equal to it. If the gate
                         is a class, it allows all instances of this class.  Default is (CNOT, Swap).
    Raises:
        TypeError: If input is for the gates is not "any" or a tuple.

    Returns:
        A list of suitable compiler engines.
    """
    if two_qubit_gates != "any" and not isinstance(two_qubit_gates, tuple):
        raise TypeError(
            "two_qubit_gates parameter must be 'any' or a tuple. When supplying only one gate, make sure to"
            "correctly create the tuple (don't miss the comma), e.g. two_qubit_gates=(CNOT,)"
        )
    if one_qubit_gates != "any" and not isinstance(one_qubit_gates, tuple):
        raise TypeError("one_qubit_gates parameter must be 'any' or a tuple.")

    rule_set = DecompositionRuleSet(modules=[projectq.libs.math, projectq.setups.decompositions])
    allowed_gate_classes = []
    allowed_gate_instances = []
    if one_qubit_gates != "any":
        for gate in one_qubit_gates:
            if inspect.isclass(gate):
                allowed_gate_classes.append(gate)
            else:
                allowed_gate_instances.append((gate, 0))
    if two_qubit_gates != "any":
        for gate in two_qubit_gates:
            if inspect.isclass(gate):
                #  Controlled gate classes don't yet exists and would require
                #  separate treatment
                if isinstance(gate, ControlledGate):  # pragma: no cover
                    raise RuntimeError('Support for controlled gate not implemented!')
                allowed_gate_classes.append(gate)
            else:
                if isinstance(gate, ControlledGate):
                    allowed_gate_instances.append((gate._gate, gate._n))  # pylint: disable=protected-access
                else:
                    allowed_gate_instances.append((gate, 0))
    allowed_gate_classes = tuple(allowed_gate_classes)
    allowed_gate_instances = tuple(allowed_gate_instances)

    def low_level_gates(eng, cmd):  # pylint: disable=unused-argument
        all_qubits = [q for qr in cmd.all_qubits for q in qr]
        if len(all_qubits) > 2:  # pragma: no cover
            raise ValueError('Filter function cannot handle gates with more than 2 qubits!')
        if isinstance(cmd.gate, ClassicalInstructionGate):
            # This is required to allow Measure, Allocate, Deallocate, Flush
            return True
        if one_qubit_gates == "any" and len(all_qubits) == 1:
            return True
        if two_qubit_gates == "any" and len(all_qubits) == 2:
            return True
        if isinstance(cmd.gate, allowed_gate_classes):
            return True
        if (cmd.gate, len(cmd.control_qubits)) in allowed_gate_instances:
            return True
        return False

    return [
        AutoReplacer(rule_set),
        TagRemover(),
        InstructionFilter(high_level_gates),
        LocalOptimizer(5),
        AutoReplacer(rule_set),
        TagRemover(),
        InstructionFilter(one_and_two_qubit_gates),
        LocalOptimizer(5),
        mapper,
        AutoReplacer(rule_set),
        TagRemover(),
        InstructionFilter(low_level_gates),
        LocalOptimizer(5),
    ]



# -*- coding: utf-8 -*-
#   Copyright 2020, 2021 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Tests for projectq.backends._aqt._aqt.py."""

import pytest
import math

from projectq import MainEngine
from projectq.backends._aqt import _aqt
from projectq.types import WeakQubitRef
from projectq.cengines import DummyEngine, BasicMapperEngine
from projectq.ops import (
    All,
    Allocate,
    Barrier,
    Command,
    Deallocate,
    Entangle,
    Measure,
    NOT,
    Rx,
    Ry,
    Rz,
    Rxx,
    S,
    Sdag,
    T,
    Tdag,
    X,
    Y,
    Z,
)


# Insure that no HTTP request can be made in all tests in this module
@pytest.fixture(autouse=True)
def no_requests(monkeypatch):
    monkeypatch.delattr("requests.sessions.Session.request")


@pytest.mark.parametrize(
    "single_qubit_gate, is_available",
    [
        (X, False),
        (Y, False),
        (Z, False),
        (T, False),
        (Tdag, False),
        (S, False),
        (Sdag, False),
        (Allocate, True),
        (Deallocate, True),
        (Measure, True),
        (NOT, False),
        (Rx(0.5), True),
        (Ry(0.5), True),
        (Rz(0.5), False),
        (Rxx(0.5), True),
        (Barrier, True),
        (Entangle, False),
    ],
)
def test_aqt_backend_is_available(single_qubit_gate, is_available):
    eng = MainEngine(backend=DummyEngine(), engine_list=[DummyEngine()])
    qubit1 = eng.allocate_qubit()
    aqt_backend = _aqt.AQTBackend()
    cmd = Command(eng, single_qubit_gate, (qubit1,))
    assert aqt_backend.is_available(cmd) == is_available


@pytest.mark.parametrize("num_ctrl_qubits, is_available", [(0, True), (1, False), (2, False), (3, False)])
def test_aqt_backend_is_available_control_not(num_ctrl_qubits, is_available):
    eng = MainEngine(backend=DummyEngine(), engine_list=[DummyEngine()])
    qubit1 = eng.allocate_qubit()
    qureg = eng.allocate_qureg(num_ctrl_qubits)
    aqt_backend = _aqt.AQTBackend()
    cmd = Command(eng, Rx(0.5), (qubit1,), controls=qureg)
    assert aqt_backend.is_available(cmd) == is_available
    cmd = Command(eng, Rxx(0.5), (qubit1,), controls=qureg)
    assert aqt_backend.is_available(cmd) == is_available


def test_aqt_backend_init():
    backend = _aqt.AQTBackend(verbose=True, use_hardware=True)
    assert len(backend._circuit) == 0


def test_aqt_empty_circuit():
    backend = _aqt.AQTBackend(verbose=True)
    eng = MainEngine(backend=backend)
    eng.flush()


def test_aqt_invalid_command():
    backend = _aqt.AQTBackend(verbose=True)

    qb = WeakQubitRef(None, 1)
    cmd = Command(None, gate=S, qubits=[(qb,)])
    with pytest.raises(Exception):
        backend.receive([cmd])


def test_aqt_sent_error(monkeypatch):
    # patch send
    def mock_send(*args, **kwargs):
        raise TypeError

    monkeypatch.setattr(_aqt, "send", mock_send)

    backend = _aqt.AQTBackend(verbose=True)
    eng = MainEngine(backend=backend)
    qubit = eng.allocate_qubit()
    Rx(0.5) | qubit
    with pytest.raises(Exception):
        qubit[0].__del__()
        eng.flush()
    # atexit sends another FlushGate, therefore we remove the backend:
    dummy = DummyEngine()
    dummy.is_last_engine = True
    eng.next_engine = dummy


def test_aqt_too_many_runs():
    backend = _aqt.AQTBackend(num_runs=300, verbose=True)
    eng = MainEngine(backend=backend, engine_list=[])
    with pytest.raises(Exception):
        qubit = eng.allocate_qubit()
        Rx(math.pi / 2) | qubit
        eng.flush()

    # Avoid exception at deletion
    backend._num_runs = 1
    backend._circuit = []


def test_aqt_retrieve(monkeypatch):
    # patch send
    def mock_retrieve(*args, **kwargs):
        return [0, 6, 0, 6, 0, 0, 0, 6, 0, 6]

    monkeypatch.setattr(_aqt, "retrieve", mock_retrieve)
    backend = _aqt.AQTBackend(retrieve_execution="a3877d18-314f-46c9-86e7-316bc4dbe968", verbose=True)

    eng = MainEngine(backend=backend, engine_list=[])
    unused_qubit = eng.allocate_qubit()
    qureg = eng.allocate_qureg(2)
    # entangle the qureg
    Ry(math.pi / 2) | qureg[0]
    Rx(math.pi / 2) | qureg[0]
    Rx(math.pi / 2) | qureg[0]
    Ry(math.pi / 2) | qureg[0]
    Rxx(math.pi / 2) | (qureg[0], qureg[1])
    Rx(7 * math.pi / 2) | qureg[0]
    Ry(7 * math.pi / 2) | qureg[0]
    Rx(7 * math.pi / 2) | qureg[1]
    del unused_qubit
    # measure; should be all-0 or all-1
    All(Measure) | qureg
    # run the circuit
    eng.flush()
    prob_dict = eng.backend.get_probabilities([qureg[0], qureg[1]])
    assert prob_dict['11'] == pytest.approx(0.4)
    assert prob_dict['00'] == pytest.approx(0.6)

    # Unknown qubit and no mapper
    invalid_qubit = [WeakQubitRef(eng, 10)]
    with pytest.raises(RuntimeError):
        eng.backend.get_probabilities(invalid_qubit)


def test_aqt_backend_functional_test(monkeypatch):
    correct_info = {
        'circuit': '[["Y", 0.5, [1]], ["X", 0.5, [1]], ["X", 0.5, [1]], '
        '["Y", 0.5, [1]], ["MS", 0.5, [1, 2]], ["X", 3.5, [1]], '
        '["Y", 3.5, [1]], ["X", 3.5, [2]]]',
        'nq': 3,
        'shots': 10,
        'backend': {'name': 'simulator'},
    }

    def mock_send(*args, **kwargs):
        assert args[0] == correct_info
        return [0, 6, 0, 6, 0, 0, 0, 6, 0, 6]

    monkeypatch.setattr(_aqt, "send", mock_send)

    backend = _aqt.AQTBackend(verbose=True, num_runs=10)
    # no circuit has been executed -> raises exception
    with pytest.raises(RuntimeError):
        backend.get_probabilities([])

    mapper = BasicMapperEngine()
    res = dict()
    for i in range(4):
        res[i] = i
    mapper.current_mapping = res

    eng = MainEngine(backend=backend, engine_list=[mapper])

    unused_qubit = eng.allocate_qubit()
    qureg = eng.allocate_qureg(2)
    # entangle the qureg
    Ry(math.pi / 2) | qureg[0]
    Rx(math.pi / 2) | qureg[0]
    Rx(math.pi / 2) | qureg[0]
    Ry(math.pi / 2) | qureg[0]
    Rxx(math.pi / 2) | (qureg[0], qureg[1])
    Rx(7 * math.pi / 2) | qureg[0]
    Ry(7 * math.pi / 2) | qureg[0]
    Rx(7 * math.pi / 2) | qureg[1]
    All(Barrier) | qureg
    del unused_qubit
    # measure; should be all-0 or all-1
    All(Measure) | qureg
    # run the circuit
    eng.flush()
    prob_dict = eng.backend.get_probabilities([qureg[0], qureg[1]])
    assert prob_dict['11'] == pytest.approx(0.4)
    assert prob_dict['00'] == pytest.approx(0.6)

    # Unknown qubit and no mapper
    invalid_qubit = [WeakQubitRef(eng, 10)]
    with pytest.raises(RuntimeError):
        eng.backend.get_probabilities(invalid_qubit)

    with pytest.raises(RuntimeError):
        eng.backend.get_probabilities(eng.allocate_qubit())



# -*- coding: utf-8 -*-
#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""
Contains a TagRemover engine, which removes temporary command tags (such as Compute/Uncompute), thus enabling
optimization across meta statements (loops after unrolling, compute/uncompute, ...)
"""
from projectq.meta import ComputeTag, UncomputeTag

from ._basics import BasicEngine


class TagRemover(BasicEngine):
    """
    TagRemover is a compiler engine which removes temporary command tags (see the tag classes such as LoopTag in
    projectq.meta._loop).

    Removing tags is important (after having handled them if necessary) in order to enable optimizations across
    meta-function boundaries (compute/ action/uncompute or loops after unrolling)
    """

    def __init__(self, tags=None):
        """
        Construct the TagRemover.

        Args:
            tags: A list of meta tag classes (e.g., [ComputeTag, UncomputeTag])
                denoting the tags to remove
        """
        super().__init__()
        if not tags:
            self._tags = [ComputeTag, UncomputeTag]
        elif isinstance(tags, list):
            self._tags = tags
        else:
            raise TypeError('tags should be a list! Got: {}'.format(tags))

    def receive(self, command_list):
        """
        Receive a list of commands from the previous engine, remove all tags which are an instance of at least one of
        the meta tags provided in the constructor, and then send them on to the next compiler engine.

        Args:
            command_list (list<Command>): List of commands to receive and then (after removing tags) send on.
        """
        for cmd in command_list:
            for tag in self._tags:
                cmd.tags = [t for t in cmd.tags if not isinstance(t, tag)]
            self.send([cmd])



# -*- coding: utf-8 -*-
#   Copyright 2021 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ==============================================================================
# This file contains:
#
# - Helper fixtures:
#   * arntask
#   * creds
#   * s3_folder
#   * device_value
#   * search_value
#   * completed_value
# - Setup fixtures for specific tests:
#   * sent_error_setup
#   * retrieve_setup
#   * functional_setup
# ==============================================================================

from io import StringIO
import json
import pytest

try:
    from botocore.response import StreamingBody
except ImportError:

    class StreamingBody:
        def __init__(self, raw_stream, content_length):
            pass


# ==============================================================================


@pytest.fixture
def arntask():
    return 'arn:aws:braket:us-east-1:id:retrieve_execution'


@pytest.fixture
def creds():
    return {
        'AWS_ACCESS_KEY_ID': 'aws_access_key_id',
        'AWS_SECRET_KEY': 'aws_secret_key',
    }


@pytest.fixture
def s3_folder():
    return ['S3Bucket', 'S3Directory']


@pytest.fixture
def device_value():
    device_value_devicecapabilities = json.dumps(
        {
            "braketSchemaHeader": {
                "name": "braket.device_schema.rigetti.rigetti_device_capabilities",
                "version": "1",
            },
            "service": {
                "executionWindows": [
                    {
                        "executionDay": "Everyday",
                        "windowStartHour": "11:00",
                        "windowEndHour": "12:00",
                    }
                ],
                "shotsRange": [1, 10],
                "deviceLocation": "us-east-1",
            },
            "action": {
                "braket.ir.jaqcd.program": {
                    "actionType": "braket.ir.jaqcd.program",
                    "version": ["1"],
                    "supportedOperations": ["H"],
                }
            },
            "paradigm": {
                "qubitCount": 30,
                "nativeGateSet": ["ccnot", "cy"],
                "connectivity": {
                    "fullyConnected": False,
                    "connectivityGraph": {"1": ["2", "3"]},
                },
            },
            "deviceParameters": {
                "properties": {
                    "braketSchemaHeader": {
                        "const": {
                            "name": "braket.device_schema.rigetti.rigetti_device_parameters",
                            "version": "1",
                        }
                    }
                },
                "definitions": {
                    "GateModelParameters": {
                        "properties": {
                            "braketSchemaHeader": {
                                "const": {
                                    "name": "braket.device_schema.gate_model_parameters",
                                    "version": "1",
                                }
                            }
                        }
                    }
                },
            },
        }
    )

    return {
        "deviceName": "Aspen-8",
        "deviceType": "QPU",
        "providerName": "provider1",
        "deviceStatus": "OFFLINE",
        "deviceCapabilities": device_value_devicecapabilities,
    }


@pytest.fixture
def search_value():
    return {
        "devices": [
            {
                "deviceArn": "arn1",
                "deviceName": "name1",
                "deviceType": "SIMULATOR",
                "deviceStatus": "ONLINE",
                "providerName": "pname1",
            },
            {
                "deviceArn": "arn2",
                "deviceName": "name2",
                "deviceType": "QPU",
                "deviceStatus": "OFFLINE",
                "providerName": "pname1",
            },
            {
                "deviceArn": "arn3",
                "deviceName": "Aspen-8",
                "deviceType": "QPU",
                "deviceStatus": "ONLINE",
                "providerName": "pname2",
            },
        ]
    }


@pytest.fixture
def completed_value():
    return {
        'deviceArn': 'arndevice',
        'deviceParameters': 'parameters',
        'outputS3Bucket': 'amazon-braket-bucket',
        'outputS3Directory': 'complete/directory',
        'quantumTaskArn': 'arntask',
        'shots': 123,
        'status': 'COMPLETED',
        'tags': {'tagkey': 'tagvalue'},
    }


# ==============================================================================


@pytest.fixture
def sent_error_setup(creds, s3_folder, device_value, search_value):

    return creds, s3_folder, search_value, device_value


@pytest.fixture
def results_json():
    return json.dumps(
        {
            "braketSchemaHeader": {
                "name": "braket.task_result.gate_model_task_result",
                "version": "1",
            },
            "measurementProbabilities": {
                "0000": 0.04,
                "0010": 0.06,
                "0110": 0.2,
                "0001": 0.3,
                "1001": 0.5,
            },
            "measuredQubits": [0, 1, 2],
        }
    )


@pytest.fixture
def retrieve_setup(arntask, creds, device_value, completed_value, results_json):

    body = StreamingBody(StringIO(results_json), len(results_json))

    results_dict = {
        'ResponseMetadata': {
            'RequestId': 'CF4CAA48CC18836C',
            'HTTPHeaders': {},
        },
        'Body': body,
    }

    return arntask, creds, completed_value, device_value, results_dict


@pytest.fixture
def functional_setup(arntask, creds, s3_folder, search_value, device_value, completed_value, results_json):
    qtarntask = {'quantumTaskArn': arntask}
    body2 = StreamingBody(StringIO(results_json), len(results_json))
    results_dict = {
        'ResponseMetadata': {
            'RequestId': 'CF4CAA48CC18836C',
            'HTTPHeaders': {},
        },
        'Body': body2,
    }

    return (
        creds,
        s3_folder,
        search_value,
        device_value,
        qtarntask,
        completed_value,
        results_dict,
    )


# ==============================================================================
